"""

    ファイル名：   TesterDialog.py

    処理内容：

        開発者用の設定変更ダイアログ

    Copyright (c) 2017-2024, SAXS Team, KEK-PF

"""

import sys
import os
import re
import warnings
import time
import glob

from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, Font, FileDialog, is_empty_val
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
import molass_legacy.KekLib.OurMessageBox as MessageBox
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting, set_dev_setting
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry, FileEntry
from molass_legacy.KekLib.BasicUtils import get_home_folder
from molass_legacy._MOLASS.SerialSettings import clear_settings, set_setting
from molass_legacy.KekLib.MultiMonitor import get_max_monitor
from molass_legacy.SerialAnalyzer.DataUtils import get_pytools_folder, get_next_subfolder

drive = __file__.split("\\")[0]

TEST_HOME = drive + r"\TODO\20240806\temp-reports"

def get_home_folder_for_test():
    return TEST_HOME
    # return get_home_folder()

class TesterDialog( Dialog ):
    def __init__( self, parent, title ):
        self.grab = 'local'     # used in grab_set
        self.parent             = parent
        self.title_             = title
        self.v2_analysis_folders = None
        self.applied            = False

    def show( self ):
        self.parent.config( cursor='wait' )
        self.parent.update()

        Dialog.__init__(self, self.parent, self.title_ )

        self.parent.config( cursor='' )

    def body( self, body_frame_ ):   # overrides parent class method

        body_frame = Tk.Frame( body_frame_ )
        body_frame.pack( padx=40, pady=10 )

        tk_set_icon_portable( self )

        clear_button = Tk.Button( body_frame, text="Clear Setting", command=self.clear_settings )
        clear_button.pack( anchor=Tk.NW )

        monitor = get_max_monitor()
        display_info = Tk.Label( body_frame, text="Display Info: " + str(monitor) )
        display_info.pack( anchor=Tk.W )

        # Tester Setting Frame
        test_all_frame = Tk.Frame( body_frame, bd=3, relief=Tk.RIDGE )
        test_all_frame.pack( anchor=Tk.W )

        # Data folder
        test_all_label = Tk.Label( test_all_frame, text="Test/Demo walking the entire folder: " )
        test_all_label.grid( row=0, column=0, pady=5 )

        entry_width = 70
        grid_row = 0

        data_folder = get_pytools_folder() + '/Data'
        self.test_folder = Tk.StringVar()
        self.test_folder.set( data_folder )
        folder_entry = FolderEntry( test_all_frame, textvariable=self.test_folder, width=entry_width )
        folder_entry.grid( row=grid_row, column=1 )

        grid_row += 1
        data_folder = get_pytools_folder() + '/Data_microfluidics'
        self.test_folder_mct = Tk.StringVar()
        self.test_folder_mct.set( data_folder )
        folder_entry = FolderEntry( test_all_frame, textvariable=self.test_folder_mct, width=entry_width )
        folder_entry.grid( row=grid_row, column=1 )

        # Result folder
        grid_row += 1
        label = Tk.Label( test_all_frame, text="Result folder: " )
        label.grid( row=grid_row, column=0, pady=5, sticky=Tk.E )
        self.result_folder = Tk.StringVar()
        self.result_folder.set(get_home_folder_for_test())
        folder_entry = FolderEntry( test_all_frame, textvariable=self.result_folder, width=entry_width )
        folder_entry.grid( row=grid_row, column=1 )

        # V2 result folder
        grid_row += 1
        test_all_label = Tk.Label( test_all_frame, text="V2 report folder: " )
        test_all_label.grid( row=grid_row, column=0, pady=5, sticky=Tk.E )
        self.v2_report_folder = Tk.StringVar()
        folder_entry = FolderEntry( test_all_frame, textvariable=self.v2_report_folder, width=entry_width )
        folder_entry.grid( row=grid_row, column=1 )
        self.v2_report_folder.trace("w", self.v2_result_folder_tracer)

        # Data Restriction
        grid_row += 1
        data_restriction_label = Tk.Label( test_all_frame, text="Data restriction: " )
        data_restriction_label.grid( row=grid_row, column=0, sticky=Tk.E)

        self.minimum_only = Tk.IntVar()
        self.minimum_only.set( 0 )
        minimum_only_cb = Tk.Checkbutton( test_all_frame, text="minimum only",
                                variable=self.minimum_only )
        minimum_only_cb.grid( row=grid_row, column=1, sticky=Tk.W )

        # Dataset Filter
        grid_row += 1
        dataset_filter_label = Tk.Label( test_all_frame, text="Dataset filter: " )
        dataset_filter_label.grid( row=grid_row, column=0, sticky=Tk.E)

        self.filter_book = Tk.StringVar()
        # book_path = os.environ["USERPROFILE"] + r"\Dropbox\_MOLASS\論文\analysis_results\selection.xlsx"
        # self.filter_book.set(book_path)
        self.filter_book_entry = FileEntry(test_all_frame, textvariable=self.filter_book, width=entry_width)
        self.filter_book_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        # Test patterns
        grid_row += 1
        test_patterns_label = Tk.Label( test_all_frame, text="Test patterns: " )
        test_patterns_label.grid( row=grid_row, column=0, sticky=Tk.E)

        test_patterns_frame = Tk.Frame( test_all_frame )
        test_patterns_frame.grid( row=grid_row, column=1, sticky=Tk.W)

        restart_str, restart_repo_dir, test_pattern, compare_folder = self.get_restart_info()
        if restart_str is None:
            restart_str = ""

        if test_pattern is None:
            test_pattern = '5'

        self.test_patterns = Tk.StringVar()
        self.test_patterns.set( '[' + test_pattern + ']')
        self.test_patterns_entry = Tk.Entry( test_patterns_frame, textvariable=self.test_patterns, width=10 )
        self.test_patterns_entry.grid( row=0, column=0, sticky=Tk.W)

        self.mapping_only = Tk.IntVar()
        self.mapping_only.set( 1 )
        mapping_only_cb = Tk.Checkbutton( test_patterns_frame, text="mapping only",
                                variable=self.mapping_only )
        mapping_only_cb.grid( row=0, column=1 )

        self.debug_trial = Tk.IntVar()
        self.debug_trial.set( 0 )
        debug_trial_cb = Tk.Checkbutton( test_patterns_frame, text="debug trial",
                                variable=self.debug_trial )
        debug_trial_cb.grid( row=0, column=2 )

        # Outline Options
        grid_row += 1
        outline_options_label = Tk.Label( test_all_frame, text='Outline Options: ' )
        outline_options_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.save_outline_figures = Tk.IntVar()
        self.save_outline_figures.set( 1 )

        cb = Tk.Checkbutton( test_all_frame, text="save outline figures",
                                variable=self.save_outline_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W )

        grid_row += 1
        self.save_3d_figures = Tk.IntVar()
        self.save_3d_figures.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save 3d figures",
                                variable=self.save_3d_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        grid_row += 1
        self.save_baseline_figures = Tk.IntVar()
        self.save_baseline_figures.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save baseline figures",
                                variable=self.save_baseline_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        # Mapping Options
        grid_row += 1
        mapping_options_label = Tk.Label( test_all_frame, text='Mapping Options: ' )
        mapping_options_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.save_mapping_figures = Tk.IntVar()
        self.save_mapping_figures.set( 1 )

        cb = Tk.Checkbutton( test_all_frame, text="save mapping figures",
                                variable=self.save_mapping_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        grid_row += 1
        self.save_cdi_figures = Tk.IntVar()
        self.save_cdi_figures.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save cdi figures",
                                variable=self.save_cdi_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        grid_row += 1
        self.save_decomp_figures = Tk.IntVar()
        self.save_decomp_figures.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save decomposition figures",
                                variable=self.save_decomp_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        # Preview Options
        grid_row += 1
        preview_options_label = Tk.Label( test_all_frame, text='Preview Options: ' )
        preview_options_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.save_preview_figures = Tk.IntVar()
        self.save_preview_figures.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save preview figures",
                                variable=self.save_preview_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        grid_row += 1
        self.save_preview_results = Tk.IntVar()
        self.save_preview_results.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save preview results",
                                variable=self.save_preview_results  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        grid_row += 1
        self.save_preview_cdi_figures = Tk.IntVar()
        self.save_preview_cdi_figures.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save preview cdi figures",
                                variable=self.save_preview_cdi_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        # V2 Options
        grid_row += 1
        v2_options_label = Tk.Label( test_all_frame, text='V2 Options: ' )
        v2_options_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.save_peakeditor_figures = Tk.IntVar()
        self.save_peakeditor_figures.set( 0 )

        cb = Tk.Checkbutton( test_all_frame, text="save peakeditor figures",
                                variable=self.save_peakeditor_figures  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )

        # Restart file string
        grid_row +=1
        start_infolder_string_label = Tk.Label( test_all_frame, text="Restart file string: " )
        start_infolder_string_label.grid( row=grid_row, column=0, sticky=Tk.E)

        self.start_infolder_string = Tk.StringVar()
        if restart_str is not None:
            self.start_infolder_string.set( restart_str )
        self.start_infolder_string_entry = Tk.Entry( test_all_frame, textvariable=self.start_infolder_string, width=20 )
        self.start_infolder_string_entry.grid( row=grid_row, column=1, sticky=Tk.W)

        # Restart report-subfolder
        grid_row +=1
        report_subfolder_label = Tk.Label( test_all_frame, text="Restart report-subfolder: " )
        report_subfolder_label.grid( row=grid_row, column=0, sticky=Tk.E)

        self.report_subfolder = Tk.StringVar()
        if restart_repo_dir is not None:
            self.report_subfolder.set( restart_repo_dir )
        self.report_subfolder_entry = Tk.Entry( test_all_frame, textvariable=self.report_subfolder, width=20 )
        self.report_subfolder_entry.grid( row=grid_row, column=1, sticky=Tk.W)

        grid_row +=1
        guinier_propgram_label = Tk.Label( test_all_frame, text="Guinier analysis program: " )
        guinier_propgram_label.grid( row=grid_row, column=0, sticky=Tk.E)

        guinier_propgram_frame = Tk.Frame( test_all_frame )
        guinier_propgram_frame.grid( row=grid_row, column=1, sticky=Tk.W)

        self.use_simpleguinier = Tk.IntVar()
        self.use_simpleguinier.set( get_dev_setting( 'use_simpleguinier' ) )
        for i, pname in enumerate([ 'SimpleGuinier', 'GP-Guinier' ]):
            rb = Tk.Radiobutton( guinier_propgram_frame, text=pname,
                        variable=self.use_simpleguinier, value=1-i,
                        )
            rb.grid( row=0, column=i, sticky=Tk.W )

        # Constant Term in Regression
        grid_row += 1
        extrapolation_options_label = Tk.Label( test_all_frame, text='Extrapolation Options: ' )
        extrapolation_options_label.grid( row=grid_row, column=0, sticky=Tk.E )

        # save extrapolated data
        self.tester_zx_save = Tk.IntVar()
        self.tester_zx_save.set( get_dev_setting( 'tester_zx_save' ) )

        zx_out_folder_cb = Tk.Checkbutton( test_all_frame,
                                        text='save extrapolated data',
                                        variable=self.tester_zx_save,
                                        state=Tk.NORMAL )
        zx_out_folder_cb.grid( row=grid_row, column=1, sticky=Tk.W )

        # 
        grid_row += 1
        shutdown_options_label = Tk.Label( test_all_frame, text='Shutdown Options: ' )
        shutdown_options_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.shutdown_machine = Tk.IntVar()
        self.shutdown_machine.set(0)

        cb = Tk.Checkbutton( test_all_frame, text="power off the machine when finished",
                                variable=self.shutdown_machine  )
        cb.grid( row=grid_row, column=1, sticky=Tk.W  )


        # Report folder to compare with
        grid_row += 1
        to_compare_label = Tk.Label( test_all_frame, text="Report folder to compare with: " )
        to_compare_label.grid( row=grid_row, column=0, pady=5 )

        self.to_compare_folder = Tk.StringVar()
        if compare_folder is None:
            compare_folder = ''
        self.to_compare_folder.set( compare_folder )
        folder_entry = FolderEntry( test_all_frame, textvariable=self.to_compare_folder, width=entry_width )
        folder_entry.grid( row=grid_row, column=1 )

        # global grab cannot be set befor windows is 'viewable'
        # and this happen in mainloop after this function returns
        # Thus, it is needed to delay grab setting of an interval
        # long enough to make sure that the window has been made
        # 'viewable'
        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

        self.update()

    def previous_results_exist(self):
        reports_folder = self.home + '/reports'
        if os.path.exists(reports_folder):
            nodes = os.listdir(reports_folder)
            ret = len(nodes) > 0
        else:
            ret = False
        return ret

    def rename_previous_results(self):
        ext_re = re.compile(r"-?(\d*)$")
        def new_postfix(m):
            # print('new_postfix: ', m.group(1))
            if m.group(1) == '':
                n = 0
            else:
                n = int(m.group(1))
            return '-' + str(n+1)

        def get_new_folder_name(folder):
            folder_ = folder
            while os.path.exists(folder_):
                folder_ = re.sub(ext_re, new_postfix, folder_, 1)   # the 4-th arg is required to let it replace only once
            return folder_

        for name in ['reports', 'images-outline', 'images-baseline',
                        'images-3d', 'images-mapping',
                        'images-cdi',
                        'images-decomp-uv', 'images-decomp-xray', 'images-preview',
                        'images-preview-cdi',
                        'images-peakeditor',
                        'qmm-images',
                        'images-fullopt_init']: # TODO: unifiy these names with those in Tester.py
            folder = os.path.join(self.home, name)
            if os.path.exists(folder):
                new_folder = get_new_folder_name(folder)
                os.rename(folder, new_folder)

    def get_last_report_folder(self):
        repo_folder = os.path.join(self.home, 'reports')
        n = 1
        while os.path.exists(repo_folder + '-' + str(n)):
            n += 1
        n -= 1
        return repo_folder + '-' + str(n)

    def to_be_conitued_reports_dir_exists( self ):
        self.to_be_continued_reports_dir = self.get_last_report_folder()
        return os.path.exists( self.to_be_continued_reports_dir )

    def get_restart_info( self ):
        self.home = get_home_folder_for_test()

        if self.previous_results_exist():
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            ok = MessageBox.askyesno('Previous Results',
                                ("Previous results seem to exist.\n"
                                "Would you like to rename them?"),
                                parent=self)
            if ok:
                self.rename_previous_results()
                MessageBox.showinfo('Previous Results',
                                "Previous results are successfully renamed.",
                                parent=self)

        self.restart = self.to_be_conitued_reports_dir_exists()
        if not self.restart:
            return [ None ] * 4

        repo_dirs = os.listdir( self.to_be_continued_reports_dir )
        # print( 'repo_dirs=', repo_dirs )
        last_successful_dir = self.to_be_continued_reports_dir + '/' + repo_dirs[-2]
        print( 'last_successful_dir=', last_successful_dir )
        last_successful_log = last_successful_dir + '/molass.log'
        start_line_re = re.compile( r'start analysis for (.+) with test pattern (\d+)' )
        fh = open( last_successful_log )
        restart_str, restart_repo_dir = None, None
        test_pattern = ''
        for line in fh:
            m = start_line_re.search( line )
            if m:
                next_subfolder = get_next_subfolder( self.test_folder.get(), m.group(1) )
                restart_str = '/'.join( next_subfolder.split( '/' )[-2:] )
                restart_repo_dir = repo_dirs[-1]
                test_pattern = m.group(2)
                break

        if test_pattern == '':
            restart_str, restart_repo_dir, test_pattern = self.get_restart_info_v2(last_successful_log, repo_dirs)

        fh.close()
        compare_folder = self.get_previous_compare_folder()
        print( 'compare_folder=', compare_folder )

        return restart_str, restart_repo_dir, test_pattern, compare_folder

    def get_restart_info_v2(self, last_successful_log, repo_dirs):
        start_line_re = re.compile( r"started loading (\S+)" )
        fh = open( last_successful_log )
        restart_str, restart_repo_dir = None, None
        test_pattern = ''
        for line in fh:
            m = start_line_re.search(line)
            if m:
                next_subfolder = get_next_subfolder( self.test_folder.get(), m.group(1) )
                restart_str = '/'.join( next_subfolder.split( '/' )[-2:] )
                restart_repo_dir = repo_dirs[-1]
                test_pattern = "20"
                break
        fh.close()
        return restart_str, restart_repo_dir, test_pattern

    def get_previous_compare_folder( self ):
        folder = None
        if False:
            # get_log_path was removed
            from molass_legacy.Test.TesterLogger import get_log_path
            log_path = get_log_path()
            if os.path.exists( log_path ):
                fh = open( log_path )
                previous_start_line = fh.readline()
                print( 'previous_start_line=', previous_start_line )
                fh.close()
                compare_folder_re = re.compile( r'with compare folder (.+)$' )
                m = compare_folder_re.search( previous_start_line )
                if m:
                    folder = m.group(1)
        return folder

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Run", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def clear_settings( self ):
        ok = MessageBox.askyesno( 'Warninig', 'This operations is not cancelable. OK?' )
        if ok:
            clear_settings()

    def v2_result_folder_tracer(self, *args):
        self.test_patterns.set("[20]")

    def validate(self):
        if is_empty_val(self.result_folder.get()):
            MessageBox.showerror('Invalid Test Patterns',
                                "Result folder must not be empty.\n",
                                parent=self)
            return 0

        try:
            test_patterns = eval(self.test_patterns.get())
            v2_report_folder = self.v2_report_folder.get()

            if (test_patterns[0] < 20 or test_patterns[0] >= 30) and is_empty_val(v2_report_folder):
                return 1

            assert not is_empty_val(v2_report_folder)
        except:
            # TODO: restore V2 result folder
            MessageBox.showerror('Invalid Test Patterns',
                                "Test Pattern must be in [20, 30) for V2 result trace\n"
                                "or V2 result folder must not be empty.",
                                parent=self)
            return 0

        try:
            self.v2_analysis_folders = glob.glob(v2_report_folder + r"\analysis-*")
            assert len(self.v2_analysis_folders) > 0
        except:
            MessageBox.showerror('Invalid V2 Result Folder',
                                "The folder does not seem to properly include V2 results.",
                                parent=self)
            return 0

        return 1

    def apply( self ):  # overrides parent class method
        self.applied    = True
        set_dev_setting( 'use_simpleguinier',       self.use_simpleguinier.get() )
        set_dev_setting( 'tester_zx_save',          self.tester_zx_save.get() )
        self.test_all()

    def test_all( self ):
        # withdraw/deiconify does not seem to be appropriate here

        # enable DebugPlot during test-execution
        from molass_legacy.KekLib.DebugPlot import set_plot_env
        set_plot_env(sub_parent=self.parent)

        set_dev_setting('running_with_tester', True)

        try:
            print( self.test_patterns.get() )
            test_patterns = eval( self.test_patterns.get() )
            assert type( test_patterns) == list, 'Not a list'
        except Exception as exc:
            print( exc )
            self.test_patterns_entry.config( fg='red' )
            MessageBox.showerror( 'Value Error', str(exc), parent=self )
            self.test_patterns_entry.focus_force()
            return

        filter_book = self.filter_book.get()
        if is_empty_val(filter_book):
            ds_filter = None
        else:
            try:
                from .DatasetFilter import DatasetFilter
                assert os.path.exists(filter_book)

                ds_filter = DatasetFilter(filter_book)
            except Exception as exc:
                print( exc )
                self.filter_book_entry.config( fg='red' )
                MessageBox.showerror( 'Value Error', str(exc), parent=self )
                self.filter_book_entry.focus_force()
                return

        self.test_patterns_entry.config( fg='black' )

        to_compare_logfolder = self.to_compare_folder.get()
        if is_empty_val( to_compare_logfolder ):
            to_compare_logfolder = None

        def test_all_closure():
            from molass_legacy.Test.Tester import Tester
            tester = Tester( self.parent )
            test_folder = self.test_folder.get()
            test_folder_mct = self.test_folder_mct.get()
            mapping_only = self.mapping_only.get() == 1
            debug_trial = self.debug_trial.get() == 1
            tester.test_all( [test_folder, test_folder_mct],
                                result_folder=self.result_folder.get(),
                                minimum_only=self.minimum_only.get(),
                                ds_filter=ds_filter,
                                mapping_only=mapping_only,
                                debug_trial=debug_trial,
                                save_outline_figures=self.save_outline_figures.get(),
                                save_3d_figures=self.save_3d_figures.get(),
                                save_baseline_figures=self.save_baseline_figures.get(),
                                save_mapping_figures=self.save_mapping_figures.get(),
                                save_cdi_figures=self.save_cdi_figures.get(),
                                save_decomp_figures=self.save_decomp_figures.get(),
                                save_preview_figures=self.save_preview_figures.get(),
                                save_preview_results=self.save_preview_results.get(),
                                save_preview_cdi_figures=self.save_preview_cdi_figures.get(),
                                save_peakeditor_figures=self.save_peakeditor_figures.get(),
                                shutdown_machine=self.shutdown_machine.get(),
                                test_patterns=test_patterns,
                                start_str=self.start_infolder_string.get(),
                                report_subfolder=self.report_subfolder.get(),
                                to_compare_logfolder=to_compare_logfolder,
                                v2_analysis_folders=self.v2_analysis_folders,
                                )

        self.parent.after( 100, test_all_closure )
        self.destroy()

class LogDiffDialog( Dialog ):
    def __init__( self, parent, title ):
        self.grab = 'local'     # used in grab_set
        self.parent             = parent
        self.title_             = title
        self.applied            = False

    def show( self ):
        self.parent.config( cursor='wait' )
        self.parent.update()

        Dialog.__init__(self, self.parent, self.title_ )

        self.parent.config( cursor='' )

    def body( self, body_frame_ ):   # overrides parent class method

        body_frame = Tk.Frame( body_frame_ )
        body_frame.pack( padx=40, pady=10 )

        tk_set_icon_portable( self )

        comming_label = Tk.Label( body_frame, text='Comming soon' )
        comming_label.pack()



"""
    ColumnInterpGPU.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
import cupy as np

class ColumnInterp:
    def __init__(self, D, j0=None):
        # add same columns to both sides
        self.D_ = np.vstack([D[:,0], D.T, D[:,-1]]).T
        self.size = D.shape[1]
        self.j0 = j0

    def safe_index(self, i):
        i_ = np.array(i, dtype=int)
        return np.max(np.array([np.zeros(i.shape, dtype=int),
                        np.min(np.array([np.ones(i.shape, dtype=int)*(self.size+1), i_]), axis=0)]), axis=0)

    def __call__(self, j):
        if type(j) != np.ndarray:
            j = np.array(j)
        if self.j0 is None:
            self.j0 = int(j[0])
        j_ = j - self.j0 + 1
        col, res = np.divmod(j_, np.ones(len(j)))
        col1 = self.safe_index(col)
        col2 = self.safe_index(col+1)
        return self.D_[:,col1]*(1-res) + self.D_[:,col2]*res



"""
    EdViewer.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.OurToplevel import OurToplevel
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import get_color, MplBackGround, reset_to_default_style
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from .EdPlotter import ed_scatter

class EdViewer(OurToplevel):
    def __init__(self, parent, path=None, data=None, save_button=False, on_close=None):
        self.mpl_bg = MplBackGround()
        self.parent = parent
        self.path = path
        self.save_button = save_button
        self.data = data
        self.on_close = on_close
        OurToplevel.__init__(self, parent, "Electron Density Viewer")

    def close(self):
        self.destroy()
        if self.on_close is not None:
            self.on_close()

    def show(self):
        # dummy for compatibility with Dialog
        pass

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.BOTH, expand=1)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        self.fig  = plt.figure(figsize=(16,7))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.build_control_panel(pframe)
        self.draw(self.path)

        self.add_dnd_bind()

    def add_dnd_bind(self):
        self.mpl_canvas_widget.register_drop_target("*")

        def dnd_handler(event):
            self.on_entry(event)

        self.mpl_canvas_widget.bind("<<Drop>>", dnd_handler)

    def build_control_panel(self, pframe):
        pass

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        if self.save_button:
            w = Tk.Button(box, text="Save", width=10, command=self.save)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

    def ok(self):
        self.close()

    def draw(self, file):
        fig = self.fig
        fig.clf()
        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122, aspect='equal')
        self.axes = [ax1, ax2]

        if file is None:
            self.draw_blank()
        else:
            if self.data is None:
                import mrcfile
                with mrcfile.open(file) as mrc:
                    data = mrc.data
            else:
                data = self.data
            self.esc = ed_scatter(fig, self.axes, data, file)
            # call self.esc.make_anim with after
            # to release the D&D source window
            # self.after(0, self.esc.make_anim)

        self.mpl_canvas.draw()

    def draw_blank(self):
        fig = self.fig
        for ax in self.axes:
            ax.set_axis_off()
        ax1, ax2 = self.axes
        ax2.set_xlim(0,1)
        ax2.text(0.5, 0.5, "Drag and drop\nan mrc file to view.", ha='center', fontsize=50, alpha=0.5)
        fig.tight_layout()

    def on_entry(self, event):
        files = event.data.split(' ')
        print('on_entry:', files)

        self.mpl_bg = MplBackGround()
        self.draw(files[0])
        reset_to_default_style()

    def save(self):
        pass



"""

    QuickAnalysis.JudgeHolder.py

    the purpose of this class is to separate CD-related codes from the mapper and sd

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
import logging
from bisect import bisect_right
from molass_legacy._MOLASS.SerialSettings import set_setting
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Conc.ConcDepend import CD_COLORS, CD_BOUNDARIES
from molass_legacy.QuickAnalysis.PreDecomposer import PreDecomposer

def rdr_is_uncomputable(rdr_hints, k):
    return rdr_hints is not None and not rdr_hints[k][0]

def convert_to_colors(cd_list, rdr_hints):
    colors = []
    for k, scd in enumerate(cd_list):
        if scd is None:
            color = 'black'
        else:
            color = CD_COLORS[bisect_right(CD_BOUNDARIES, scd)]

        if rdr_is_uncomputable(rdr_hints, k):
            if color == "green":
                color = 'white'
            else:
                color = 'orange'
        colors.append(color)
    return colors

def convert_to_degrees(cd_list, rdr_hints):
    degrees = []
    for k, scd in enumerate(cd_list):
        if scd is None:
            degree = 1
        else:
            cd_level = bisect_right(CD_BOUNDARIES, scd)
            degree = 2 if cd_level> 0 else 1
        degrees.append(degree)
    return degrees

class JudgeHolder:
    def __init__(self, sd, pre_recog, analysis_copy):
        self.logger = logging.getLogger(__name__)
        self.update_count = 0
        self.sd_orig = sd
        self.pre_recog = pre_recog
        self.sd = analysis_copy
        self.predec = PreDecomposer(sd, pre_recog, analysis_copy, debug=False)
        self.rank_info = None
        self.rdr_hints = None
        self.cd = None
        self.cd_info = None
        self.cd_list = None
        self.cd_colors = None
        self.cd_degrees = None
        self.guide_info = None
        self.x_curve = None

    def set_mapper(self, mapper):
        self.mapper = mapper        # used for RDR calculation

    def update_sd(self, new_sd):
        # called by mapper
        self.sd = new_sd
        self.predec.update_sd_copy(new_sd)
        self.update_count += 1

    def clear_cd_info_dependents(self):
        # called by mapper
        self.cd_list = None
        self.cd_colors = None
        self.cd_degrees = None

    def update_x_curve(self, curve):
        # called by mapper
        self.x_curve = curve

    def get_callbacks(self):
        return self.update_sd, self.clear_cd_info_dependents, self.update_x_curve

    def get_cd_list(self):
        if self.cd_list is None:
            if self.update_count == 0:
                self.logger.warning("using not-yet-corrected sd to get cd_list is not desirable.")
            peak_info = self.x_curve.peak_info
            cd_list = []
            cd_infos = self.get_conc_depend_info()
            n = len(cd_infos)
            kstart = 0
            for p, pkrec in enumerate(peak_info):
                top_x_xr = pkrec[1]
                snd = None
                for k in range(kstart, n):
                    cdrec = cd_infos[k]
                    top_x_cd = cdrec[0]
                    if abs(top_x_cd - top_x_xr) < 5:
                        snd = cdrec[1]
                        kstart = k + 1
                        break
                    else:
                        continue
                cd_list.append(snd)
            self.logger.info("got cd_list as " + str(cd_list))
            self.cd_list = cd_list
        return self.cd_list

    def add_conc_depend_info(self):
        from molass_legacy.Conc.ConcDepend import ConcDepend
        rdr_hints = self.get_rdr_hints()

        q, data, error = self.sd.get_xray_data()
        ecurve = self.mapper.x_curve
        cd = ConcDepend(q, data, error, ecurve, xray_scale=self.sd.get_xray_scale())
        self.cd_info = cd.compute_judge_info(rdr_hints, debug_info=[self])
        self.logger.info("added cd_info as %s", str(self.cd_info))

    def get_conc_depend_info(self):
        if self.cd_info is None:
            if self.sd.baseline_corrected:
                self.logger.info("safely adding cd_info: %s", self.sd.get_id_info())
            else:
                self.logger.warning("adding cd_info before baseline correction is not desirable: %s", self.sd.get_id_info())
            self.add_conc_depend_info()
        return self.cd_info

    def has_uncomputable_rdrs(self):
        if self.rdr_hints is None:
            return False

        ret_judge = False
        for computable, rdr in self.rdr_hints:
            if not computable:
                ret_judge = True
                break
        return ret_judge

    def get_rdr_hints(self):
        if self.rdr_hints is None:
            self.compute_rdr_hints()
        return self.rdr_hints

    def compute_rdr_hints(self):
        from .RgDiffRatios import RgDiffRatios
        sd = self.sd
        self.rdrs = RgDiffRatios(sd, self.mapper.x_curve)
        self.rdr_hints = self.rdrs.get_rank_hints()
        self.logger.info("rdr_hints=%s", str(self.rdr_hints))

    def get_cd_colors(self, guide_info=None, debug=False):
        """
        see ElutionMapperCanvas.get_cd_colors to understand the following control flow.
            1. self.get_cd_list() # substantial computation which implies RDR computation
            2. caller.update_guide_info()   # which adds RDR-info to guide_info
            3. convert_to_colors(cd_list, guide_info)
        """
        if guide_info is None:
            # do this ahead of getting the guild_info
            self.get_cd_list()
            return  # to let the caller.update_guide_info()
        else:
            # do the following after the caller.update_guide_info()
            self.guide_info = guide_info

        if self.cd_colors is None:
            self.logger.info("guide_info.has_an_uncomputable_rdr()? %s", str(guide_info.has_an_uncomputable_rdr()))
            cd_list = self.get_cd_list()    # this is skipping the comutation
            rdr_hints = self.get_rdr_hints()
            self.cd_colors =  convert_to_colors(cd_list, rdr_hints)
            set_setting('mapper_cd_color_info', (self.x_curve.peak_top_x.copy(), self.cd_colors.copy()))
            self.logger.info("mapper_cd_color_info=(%s, %s, %s)", str(self.x_curve.peak_top_x), str(self.cd_colors), str(rdr_hints))
            if debug:
                x_curve = self.x_curve
                x = x_curve.x
                y = x_curve.y
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("get_cd_colors debug")
                    ax.plot(x, y)
                    for px in x_curve.peak_top_x:
                        ax.plot(px, x_curve.spline(px), "o")
                    fig.tight_layout()
                    plt.show()

        return self.cd_colors

    def get_cd_degrees(self):
        if self.cd_degrees is None:
            rdr_hints = self.get_rdr_hints()
            self.cd_degrees =  convert_to_degrees(self.get_cd_list(), rdr_hints)
        return self.cd_degrees

    def get_cd_degree_from_range(self, start, stop):
        cd_degrees = self.get_cd_degrees()
        pno = self.x_curve.peak_no_from_range(start, stop)
        if pno is None:
            degree = 1
            self.logger.warning("conc. dependency is set to %d for unidentified range %d:%d", degree, start, stop)
        else:
            degree = self.cd_degrees[pno]
        return degree



"""
    Models/RateTheory/EDM_Demo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

def demo(in_folder, sd, num_peaks=None):
    from importlib import reload
    import Models.RateTheory.EDM
    reload(Models.RateTheory.EDM)
    from molass_legacy.Models.RateTheory.EDM import guess_multiple, edm_impl

    uv_ecurve = sd.get_xray_curve()
    xr_ecurve = sd.get_uv_curve()

    for ecurve in uv_ecurve, xr_ecurve:

        x = ecurve.x
        y = ecurve.y

        if num_peaks is None:
            num_peaks_ = len(ecurve.peak_info) + 1
        else:
            num_peaks_ = num_peaks

        guess_multiple(x, y, num_peaks_, debug=True)



"""
    Optimizer.RestartPatcher.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.widgets import Slider, Button
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from .FvScoreConverter import convert_score

class RestartPatcher:
    def __init__(self, caller, optimizer, work_params, jobname):
        self.caller = caller
        self.optimizer = optimizer
        self.work_params = work_params
        self.jobname = jobname

    def preview(self):
        return patch_and_restart_impl(self.caller, self.optimizer, self.work_params, self.jobname)

    def apply(self, restart_params):
        self.caller.start_next_trial(restart_params=restart_params, real_bounds=self.optimizer.real_bounds)

def patch_and_restart_from_editor(editor):
    optimizer = editor.optimizer
    init_params = editor.get_init_params()
    jobname = "---"
    rp = RestartPatcher(editor, optimizer, init_params, jobname)
    ret = rp.preview()
    if ret is not None:
        rp.apply(ret)

def patch_and_restart(caller):
    print("patch_and_restart")
    canvas = caller.canvas
    optimizer = canvas.fullopt
    init_params = canvas.get_current_params()
    job_info = canvas.dialog.get_job_info()
    jobname = job_info[0]
    rp = RestartPatcher(caller, optimizer, init_params, jobname)
    ret = rp.preview()
    if ret is not None:
        rp.apply(ret)

def patch_and_restart_impl(caller, optimizer, init_params, jobname):
    num_components = optimizer.get_num_components()
    in_folder = get_in_folder()
    bounds_array = np.array(optimizer.params_type.get_param_bounds(init_params, real_bounds=optimizer.real_bounds))     # reconsider the need here of real_bounds kwarg
    psinfo = optimizer.get_paramslider_info()
    print("psinfo.cmpparam_names", psinfo.cmpparam_names)
    work_params = init_params.copy()

    def slider_update(k, val):
        print([k], "slider_update", val)
        work_params[k] = val

    def draw_params(params, fig, axes):
        fv = optimizer.objective_func(params)
        sv = convert_score(fv)
        ax1.set_title("UV Decomposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)
        ax3.set_title("Objevive Function Scores in %.1f" % sv, fontsize=16)
        optimizer.objective_func(params, plot=True, axis_info=(fig, axes))

    def add_sliders(component, indeces, fig, slider_axes, sliders, left, width, slider_params, slider_bounds):
        print("add_sliders", component)
        slider_specs = []
        for i, name in enumerate(psinfo.cmpparam_names):
            slider_specs.append((indeces[i], name, *slider_bounds[i], slider_params[i]))

        for k, (j, label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            print([k], j, label, component+i)
            ax_ = fig.add_axes([left, 0.4 - 0.05*k, width, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=j: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

    modelname = optimizer.get_model_name()
    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,10))
        fig.suptitle("%s Parameter Inspection for Patch and Restart for Job %s in %s" % (modelname, jobname, in_folder), fontsize=20)
        axt = ax2.twinx()
        axt.grid(False)
        axes = (ax1, ax2, ax3, axt)
        draw_params(init_params, fig, axes)

        slider_axes = []
        sliders = []

        def redraw_slider_params(event):
            for ax in axes:
                ax.clear()
            draw_params(work_params, fig, axes)
            fig.canvas.draw_idle()

        def devtest_slider_params(event):
            from importlib import reload
            import Optimizer.Devel.DevMeasureInspect
            reload(Optimizer.Devel.DevMeasureInspect)
            from molass_legacy.Optimizer.Devel.DevMeasureInspect import dev_measure_inspect
            print("devtest_slider_params")
            dev_measure_inspect(optimizer, work_params)

        def try_code_patch(event):
            from importlib import reload
            import Optimizer.Devel.CodePatch
            reload(Optimizer.Devel.CodePatch)
            from molass_legacy.Optimizer.Devel.CodePatch import code_patch
            print("devtest_slider_params")
            current_folder = caller.get_curr_work_folder()      # in case of FullOptDialog
            code_patch(optimizer, work_params, current_folder)

        def reset_slider_params(event):
            nonlocal work_params
            for ax in axes:
                ax.clear()
            work_params = init_params.copy()
            draw_params(work_params, fig, axes)
            for slider in sliders:
                slider.reset()
            fig.canvas.draw_idle()

        def simple_inspect(event):
            from importlib import reload
            import Optimizer.Devel.SimpleInspect
            reload(Optimizer.Devel.SimpleInspect)
            from molass_legacy.Optimizer.Devel.SimpleInspect import simple_inspect_impl
            print("devtest_slider_params")
            simple_inspect_impl(optimizer, work_params)

        for i in range(num_components):
            width = 0.9/num_components
            left = 0.08 + width*i
            slider_width = width*0.6
            indeces = psinfo.get_component_indeces(i)
            print([i], "indeces=", indeces)
            slider_params = work_params[indeces]
            slider_bounds = bounds_array[indeces,:]
            ax = fig.add_axes([left, 0.45, width, 0.03])
            ax.set_axis_off()
            ax.text(0.3, 0.5, "component-%d" % (i+1), ha='center', va='center')
            add_sliders(i, indeces, fig, slider_axes, sliders, left, slider_width, slider_params, slider_bounds)

        for i, (name, j) in enumerate(zip(psinfo.whlparam_names, psinfo.whlparam_indeces)):
            ax_ = fig.add_axes([0.08, 0.2 - 0.05*i, 0.1, 0.03])
            ax_.set_axis_off()
            valmin = bounds_array[j,0]
            valmax = bounds_array[j,1]
            valinit = work_params[j]
            slider  = Slider(ax_, label=name, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=j: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        ax_redraw = fig.add_axes([0.4, 0.05, 0.1, 0.05])
        redraw_button = Button(ax_redraw, 'Redraw')
        redraw_button.on_clicked(redraw_slider_params)
        ax_devtest = fig.add_axes([0.5, 0.05, 0.1, 0.05])
        devtest_button = Button(ax_devtest, 'Devtest')
        devtest_button.on_clicked(devtest_slider_params)
        ax_devtest = fig.add_axes([0.6, 0.05, 0.1, 0.05])
        devtest_button = Button(ax_devtest, 'CodePatch')
        devtest_button.on_clicked(try_code_patch)
        ax_reset = fig.add_axes([0.7, 0.05, 0.1, 0.05])
        reset_button = Button(ax_reset, 'Reset')
        reset_button.on_clicked(reset_slider_params)
        ax_reset = fig.add_axes([0.8, 0.05, 0.1, 0.05])
        reset_button = Button(ax_reset, 'Inspect')
        reset_button.on_clicked(simple_inspect)


        fig.tight_layout()
        fig.subplots_adjust(bottom=0.55)
        ret = plt.show()

    if ret:
        return work_params
    else:
        return None


# coding: utf-8
"""
    MXD.Bayesian.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import numpy as np
from bayes_opt import BayesianOptimization
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans
from molass_legacy.Peaks.ElutionModels import egh
from .FastDecomposer import FastDecomposer
import molass_legacy.KekLib.DebugPlot as plt

def spike_demo():
    x = np.arange(500)
    y = np.zeros(len(x))
    for params in [
            # [0.6, 200, 40, 10],
            # [0.6, 200, 30, 5],
            [1.0, 300, 30, 10],
            ]:
        y += egh(x, *params)

    spike_demo_impl(x, y, num_compoments=2)



def spike_demo_impl(x, y, num_compoments=None):

    def target(h1, m1, s1, t1):
        ty = egh(x, h1, m1, s1, t1)
        r2 = np.sum((ty - y)**2)
        return -r2

    pbounds = {
        'h1': (0, 2),
        'm1': (0, 400),
        's1': (0, 50),
        't1': (0, 20),
        }

    optimizer = BayesianOptimization(
        f=target,
        pbounds=pbounds,
        random_state=1,
    )

    optimizer.maximize(
        init_points=2,
        n_iter=5,
    )

    plt.push()
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
    ax1.plot(x, y)

    fig.tight_layout()
    plt.show()
    plt.pop()

def spike_demo_real(in_folder):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_sd()
    xr_curve = sd.get_xray_curve()
    spike_demo_impl(xr_curve.x, xr_curve.y, num_compoments=3)



# coding: utf-8
"""
    GuinierProofPlot.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import sys
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.ticker      import FuncFormatter
from bisect                 import bisect_right

class GuinierProofPlot:
    def __init__( self, fit, interval, gunier, optimal_interval_q=None, title=None, ):

        x_  = interval.x
        self.gunier = gunier

        x = x_**2

        y1  = np.log( interval.y )
        y2  = np.log( fit.model( x_, sigma=0 ) )
 
        self.fig    = fig   = plt.figure( figsize=( 24, 8 ) )
        self.ax1    = ax1   = fig.add_subplot( 1, 3, 1 )
        self.ax2    = ax2   = fig.add_subplot( 1, 3, 2 )
        self.ax3    = ax3   = fig.add_subplot( 1, 3, 3 )

        # for ax in [ ax1, ax2 ]: ax.set_axis_bgcolor('lightgray')

        if title is not None:
            ax1.set_title( title )

        ax2.set_title( 'Plot for Anomaly Index' )
        ax3.set_title( 'Curvature(s)' )

        for ax in [ax1, ax2]:
            ax.plot( x, y1, 'o' )
            ax.xaxis.set_major_formatter( FuncFormatter( self.x_formatter ) )

            ax.plot( x, y2, color='pink', label='Guinier-Porod model fitted' )

            if interval.smoother is not None:
                if interval.smoother.is_guinier:
                    y3  = interval.smoother( x )
                else:
                    y3  = np.log( interval.smoother( x_ ) )
                ax.plot( x, y3, color='cyan', label='Smooth line on the interval' )

        ymin1, ymax1 = ax1.get_ylim()
        ax1.set_ylim( [ ymin1, ymax1 ] )

        for d in range(1,5):
            q = 1 / fit.Rg * np.sqrt( 3*d/2 )
            q2 = q**2
            for ax in [ ax1, ax2 ]:
                ax.plot( [ q2, q2 ], [ ymin1, ymax1 ], ':', color='orange' )

        if optimal_interval_q is not None:
            for q in optimal_interval_q:
                q2 = q**2
                for ax in [ ax1, ax2 ]:
                    ax.plot( [ q2, q2 ], [ ymin1, ymax1 ], ':', color='green' )

        if gunier.sx_array is not None:
            colors = [ 'yellow', 'red' ]
            markers = [ 'x', '' ]
            labels  = [ 'Spline for IpI Check', 'Spline for Aggregation Check' ]
            for i in range( 1 ):
                sx = gunier.sx_array[i]
                sy = gunier.sy_array[i]
                # print( i, 'len(sx)=', len(sx) )
                ax2.plot( sx, sy, marker=markers[i], color=colors[i], label=labels[i] )
                if i == 0:
                    cy = gunier.cy_array[i]
                    ax3.plot( sx, cy, color='green' )

            xmin2 = 0.0
            xmax2 = np.max( sx ) + 0.0001
            ymin2 = np.min( sy ) - 0.1
            ymax2 = np.max( sy ) + 0.1
            ax2.set_xlim( xmin2, xmax2 )
            ax2.set_ylim( ymin2, ymax2 )

            ax1.plot(   [ xmin2, xmax2, xmax2, xmin2, xmin2 ],
                        [ ymin2, ymin2, ymax2, ymax2, ymin2 ],
                        ':', color='black' )

        ax1.legend()
        ax2.legend()
        plt.tight_layout()

    def x_formatter( self, val, pos ):
        if pos is None: return ''

        if val >= 0:
            index = bisect_right( self.gunier.x, val )
            return '%g\n(%g)' % ( val, index )
        else:
            return '%g' % val

    def show( self, block=False ):
        plt.show( block=block )

    def close( self ):
        plt.close()



"""
    Optimizer.ResidualView.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.BasicUtils import ordinal_str
from DataUtils import get_in_folder
from MatrixData import simple_plot_3d

class ResidualView(Dialog):
    def __init__(self, parent, parent_dialog, state_canvas):
        self.parent = parent
        self.parent_dialog = parent_dialog
        self.state_canvas = state_canvas
        self.sd = state_canvas.optinit_info.sd
        self.x_array = state_canvas.demo_info[1]
        self.best_index = state_canvas.get_best_index()
        self.fullopt = state_canvas.fullopt
        Dialog.__init__(self, parent, "Residual View", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X, padx=20)
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT)

        self.fig = fig = plt.figure(figsize=(12, 6))
        self.draw_residual_view(fig)

        fig.tight_layout()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_residual_view(self, fig):

        job_info = self.parent_dialog.get_job_info()
        job_name = job_info[0]
        in_folder = get_in_folder()

        best_params = self.x_array[self.best_index]
        fv, score_list, Pxr, Cxr, Puv, Cuv, mapped_UvD = self.fullopt.objective_func(best_params, return_full=True)

        Ruv = Puv @ Cuv - mapped_UvD
        Rxr = Pxr @ Cxr - self.fullopt.xrD_

        ax1 = fig.add_subplot(121, projection="3d")
        ax2 = fig.add_subplot(122, projection="3d")

        # fig.tight_layout()
        fig.subplots_adjust(top=0.8)

        fig.suptitle("Residual Views of Job %s at %s local minimium on %s" % (job_name, ordinal_str(self.best_index), in_folder), fontsize=20)
        ax1.set_title("UV Residual 3D View", fontsize=16)
        ax2.set_title("XR Residual 3D View", fontsize=16)

        simple_plot_3d(ax1, Ruv, x=self.sd.lvector)
        simple_plot_3d(ax2, Rxr, x=self.sd.qvector)



"""
    KekLib.EvalUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np      # used in the eval, in cases where the value is like 'np.int64(140)'
from numpy import array, nan, inf   # used in the eval

def eval_file(filename, locals_={}, replacer=None):
    try:
        with open(filename) as fh:
            code = fh.read()
        if replacer is not None:
            code = replacer(code)
        return eval(code, globals(), locals_)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None,  'Error in file_eval', n=10)


"""

    TesterSimple.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""

from molass_legacy.KekLib.OurTkinter import Tk, Dialog

class TesterSimple(Dialog):
    def __init__(self, parent, in_folder, sd, callback, mapper=None, debug=False):
        self.in_folder = in_folder
        self.sd = sd
        self.callback = callback
        self.mapper = mapper
        self.debug = debug
        Dialog.__init__(self, parent, "TesterSimple", visible=False)

    def show( self ):
        self._show()

    def body(self, body_frame):
        button = Tk.Button(body_frame, text="Run", command=self.run)
        button.pack(padx=50, pady=30)

    def run(self):
        kw_debug = {'debug':self.debug} if self.debug else {}

        if self.mapper is None:
            self.callback(self, self.in_folder, self.sd, **kw_debug)
        else:
            self.callback(self, self.in_folder, self.sd, self.mapper, **kw_debug)

def show_dialog(in_folder, callback, add_mapper=False, analysis_copy=False, debug=False):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.KekLib.TkUtils import get_tk_root

    set_setting('in_folder', in_folder)

    sp = StandardProcedure()
    if analysis_copy:
        from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
        sd_ = sp.load_old_way(in_folder)
        pre_recog = PreliminaryRecognition(sd_)
        sd = sd_._get_analysis_copy_impl(pre_recog)
        mapper = None
    elif add_mapper:
        sp.load(in_folder)
        sd = sp.get_corrected_sd()      # this method creates a mapper
        mapper = sp.mapper
    else:
        sd = sp.load_old_way(in_folder)
        mapper = None

    root = get_tk_root()

    def show_tester():
        ts = TesterSimple(root, in_folder, sd, callback, mapper=mapper, debug=debug)
        ts.show()
        root.quit()

    root.after(0, show_tester)
    root.mainloop()
    root.destroy()



"""
    Optimizer.BoundsInspection.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from tksheet import Sheet

class BoundsInspection(Dialog):
    def __init__(self, parent, fullopt, params):
        self.parent = parent
        self.params = params
        self.param_names = fullopt.get_parameter_names()
        self.real_bounds = fullopt.real_bounds
        self.scale_shift = fullopt.scale_shift
        self.scale_slope = fullopt.scale_slope
        self.norm_params = fullopt.to_norm_params(params)
        Dialog.__init__(self, parent, "Bounds Inspection", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):

        num_columns = 7
        num_valid_rows = min(30, len(self.real_bounds))
        data_list = [ [name] + ["%g" % v for v in [param, *bpair, sshift, sslope, nparam]]
                        for name, param, bpair, sshift, sslope, nparam
                        in zip(self.param_names, self.params, self.real_bounds, self.scale_shift, self.scale_slope, self.norm_params)]

        column_width = 90
        width = column_width*num_columns + 60
        height = int(22*num_valid_rows) + 60
        self.sheet = Sheet(body_frame, width=width, height=height, data=data_list, show_selected_cells_border=False, column_width=column_width)
        self.sheet.pack()
        self.sheet.enable_bindings()



"""
    ConcMatrix.py

    Copyright (c) 2019-2023, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting

class ConcMatrix:
    def __init__(self, x, conc, conc_depend=2, paired_ranges=None, mc_vector=None,
                    conc_factor=5,  add_conc_const=False, ecurve=None, j0=0):
        self.logger = logging.getLogger(__name__)
        self.x = x
        self.conc = conc
        self.conc_depend = conc_depend
        self.cdl_list = []
        self.paired_ranges = paired_ranges
        self.mc_vector = mc_vector
        self.conc_factor = conc_factor
        self.ecurve = ecurve
        self.j0 = j0
        self.concentration_datatype = get_setting('concentration_datatype')
        self.logger.info("creating C matrix in ranges %s with concentration_datatype=%d", str(paired_ranges), self.concentration_datatype)

        if conc is None:
            if self.concentration_datatype == 1:
                C_list = self.make_from_xr_ecurve()
            elif self.concentration_datatype == 3:
                assert mc_vector is not None
                C_list = self.make_from_mc_vector()
            else:
                """
                    note that difference info between concentration_datatype 0 and 2
                    has been already included in paired_ranges
                    when they were made in DecompEditorFrame.make_range_info()
                    by selecting opt_recs or opt_recs_uv
                """
                C_list = self.make_from_sec_model()
        else:
            C_list = self.make_from_conc()

        if add_conc_const:
            C_list.append( np.ones( len(x) ) )

        self.data = np.array( C_list )
        self.logger.info("C matrix data.shape is %s", str(self.data.shape))

    def make_from_xr_ecurve(self):
        C = self.ecurve.spline(self.x)
        C_list = [C]
        if self.conc_depend > 1:
            for k in range(2, self.conc_depend+1):
                C_list += [C**k]

        self.cdl_list.append(len(C_list))

        return C_list

    def make_from_mc_vector(self):
        C = self.mc_vector      # already scaled with conc_factor
        C_list = [C]
        if self.conc_depend > 1:
            for k in range(2, self.conc_depend+1):
                C_list += [C**k]

        self.cdl_list.append(len(C_list))

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            C_ = np.array(C_list)
            fig = plt.figure(figsize=(14,7))
            ax1 = fig.add_subplot(121)
            ax2 = fig.add_subplot(122)

            ax1.set_title("Rows of C matrix", fontsize=20)
            ax1.plot(C_[0,:], label='$C[0,:]$')
            ax1.plot(C_[1,:], label='$C[1,:]=C[0,:]^2=$')

            U, s, VT = np.linalg.svd(C_)
            ax2.set_title("Singular Values of C matrix", fontsize=20)
            ax2.plot(s[0:5], ':', marker='o')
            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(0, ymax)

            ax1.legend()
            fig.tight_layout()
            plt.show()

        return C_list

    def make_from_sec_model(self, debug=False):
        C_list = []
        x = self.x

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.KekLib.OurMatplotlib import get_color
            from molass_legacy.DataStructure.AnalysisRangeInfo import shift_paired_ranges
            print('conc_factor=', self.conc_factor)
            print('paired_ranges=', self.paired_ranges, 'conc_depend=', self.conc_depend)
            j0 = self.j0
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            axt = ax.twinx()
            ax.set_title('debug for ' + str(shift_paired_ranges(-j0, self.paired_ranges)))
            ecurve = self.ecurve
            ax.plot(ecurve.x+j0, ecurve.y, color='orange')

        for k, paired_range in enumerate(self.paired_ranges):
            cy = np.zeros(len(x))

            for elm_rec in paired_range.elm_recs:
                e   = elm_rec[0]
                fnc = elm_rec[1]
                y = fnc(x)
                cy += y
                if debug:
                    color = get_color(e)
                    x_ = self.ecurve.x
                    y_ = fnc(x_)
                    ax.plot(x_+j0, y_, ':', color=color)
                    ax.plot(x +j0, y, color=color, label=str(e))

            C = cy * self.conc_factor

            if debug:
                axt.plot(x+j0, C, color='red', label=str([k]) + ' conc. scaled')

            C_rows = [C]
            if self.conc_depend > 1:
                for k in range(2, self.conc_depend+1):
                    C_rows += [C**k]

            C_list += C_rows
            self.cdl_list.append(len(C_rows))

        if debug:
            from PlotUtils import align_yaxis_np
            ymin, ymax = ax.get_ylim()
            ymint, ymaxt = axt.get_ylim()
            axt.set_ylim(ymin, ymaxt)
            ax.legend(loc='upper left')
            axt.legend(loc='upper right')
            align_yaxis_np(ax, axt)
            fig.tight_layout()
            plt.show()
            plt.pop()

        return C_list

    def make_from_conc(self):
        if self.conc_depend == 2:
            ret_conc = self.conc
        else:
            if len(self.conc) == 4:
                ret_conc = [self.conc[0], self.conc[2]]
            else:
                # work-around for 3-state model
                # TODO: do this properly
                ret_conc = self.conc
        return ret_conc



"""
    Solvers.ABC.CustomModel.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from typing import Any, Callable, Union

from pyabc.acceptor import Acceptor
from pyabc.distance import Distance
from pyabc.epsilon import Epsilon
from pyabc.parameters import Parameter

from pyabc.model import ModelResult


class CustomObjectiveModel:
    """
    General model. This is the most flexible model class, but
    also the most complicated one to use.
    This is an abstract class and not functional on its own.
    Derive concrete subclasses for actual usage.

    The individual steps

      * sample
      * summary_statistics
      * distance
      * accept

    can be overwritten.

    To use this class, at least the sample method has to be overriden.

    .. note::

        Most likely you do not want to use this class directly, but the
        :class:`FunctionModel` instead, or even just pass a plain function
        as model.

    Parameters
    ----------
    name: str, optional (default = "model")
        A descriptive name of the model. This name can simplify further
        analysis for the user as it is stored in the database.
    """

    def __init__(self, func, name: str = "Model"):
        self.func = func
        self.name = name

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, self.name)

    def sample(self, pars: Parameter):
        """
        Return a sample from the model evaluated at parameters ``pars``. This
        can be raw data, or already summarized statistics thereof.

        This method has to be implemented by any subclass.

        Parameters
        ----------
        pars: Parameter
            Dictionary of parameters.

        Returns
        -------
        sample: any
            The sampled data.
        """
        raise NotImplementedError()

    def summary_statistics(
        self, t: int, pars: Parameter, sum_stat_calculator: Callable
    ) -> ModelResult:
        """
        Sample, and then calculate the summary statistics.

        Called from within ABCSMC during the initialization process.

        Parameters
        ----------
        t: int
            Current time point.
        pars: Parameter
            Model parameters.
        sum_stat_calculator: Callable
            A function which calculates summary statistics, as passed to
            :class:`pyabc.smc.ABCSMC`.
            The user is free to use or ignore this function.

        Returns
        -------
        model_result: ModelResult
            The result with filled summary statistics.
        """
        if self.func is None:
            raw_data = self.sample(pars)
            sum_stat = sum_stat_calculator(raw_data)
        else:
            sum_stat = None
        return ModelResult(sum_stat=sum_stat)

    def distance(
        self,
        t: int,
        pars: Parameter,
        sum_stat_calculator: Callable,
        distance_calculator: Distance,
        x_0: dict,
    ) -> ModelResult:
        """
        Sample, calculate summary statistics, and then calculate the distance.

        Not required in the current implementation.

        Parameters
        ----------
        t: int
            Current time point.
        pars: Parameter
            Model parameters.
        sum_stat_calculator: Callable
            A function which calculates summary statistics, as passed to
            :class:`pyabc.smc.ABCSMC`.
            The user is free to use or ignore this function.
        distance_calculator: Callable
            A function which calculates the distance, as passed to
            :class:`pyabc.smc.ABCSMC`.
            The user is free to use or ignore this function.
        x_0: dict
            Observed summary statistics.

        Returns
        -------
        model_result: ModelResult
            The result with filled distance.
        """

        sum_stat_result = self.summary_statistics(t, pars, sum_stat_calculator)
        if self.func is None:
            distance = distance_calculator(sum_stat_result.sum_stat, x_0, t, pars)
        else:
            distance = self.func(pars)
        sum_stat_result.distance = distance

        return sum_stat_result

    def accept(
        self,
        t: int,
        pars: Parameter,
        sum_stat_calculator: Callable,
        distance_calculator: Distance,
        eps_calculator: Epsilon,
        acceptor: Acceptor,
        x_0: dict,
    ):
        """
        Sample, calculate summary statistics, calculate distance, and then
        accept or not accept a parameter.

        Called from within ABCSMC in each iteration to evaluate a parameter.


        Parameters
        ----------
        t: int
            Current time point.
        pars: Parameter
            The model parameters.
        sum_stat_calculator: Callable
            A function which calculates summary statistics.
            The user is free to use or ignore this function.
        distance_calculator: pyabc.Distance
            The distance function.
            The user is free to use or ignore this function.
        eps_calculator: pyabc.Epsilon
            The acceptance thresholds.
        acceptor: pyabc.Acceptor
            The acceptor judging whether to accept, based on distance and
            epsilon.
        x_0: dict
            The observed summary statistics.

        Returns
        -------
        model_result: ModelResult
            The result with filled accepted field.

        """
        result = self.summary_statistics(t, pars, sum_stat_calculator)
        acc_res = acceptor(
            distance_function=distance_calculator,
            eps=eps_calculator,
            x=result.sum_stat,
            x_0=x_0,
            t=t,
            par=pars,
        )
        result.distance = acc_res.distance
        result.accepted = acc_res.accept
        result.weight = acc_res.weight

        return result


class CustomObjectiveFunction(CustomObjectiveModel):
    """
    A model which is initialized with a function which generates the samples.
    For most cases this class will be adequate.
    Note that you can also pass a plain function to the ABCSMC class, which
    then gets automatically converted to a FunctionModel.

    Parameters
    ----------
    sample_function: Callable[[Parameter], Any]
        Returns the sample to be passed to the summary statistics method.
        This function as a single argument which is a Parameter.
    name: str. optional
        The name of the model. If not provided, the names if inferred from
        the function name of `sample_function`.
    """

    def __init__(
        self,
        objective_function,
        sample_function: Callable[[Parameter], Any],
        name: str = None
    ):
        self.objective_function = objective_function
        if name is None:
            # try to get the model name
            try:
                name = sample_function.__name__
            except AttributeError:
                name = sample_function.__class__.__name__
        super().__init__(name)
        self.sample_function = sample_function

    def sample(self, pars: Parameter):
        return self.sample_function(pars)

    @staticmethod
    def to_model(func, maybe_model: Union[Callable, CustomObjectiveModel]) -> CustomObjectiveModel:
        """
        Alternative constructor. Accepts either a Model instance or a
        function and returns always a Model instance.

        Parameters
        ----------
        maybe_model:
            Constructs a FunctionModel instance if a function is passed.
            If a Model instance is passed, the Model instance itself is
            returned.

        Returns
        -------
        model: A valid model instance
        """
        if isinstance(maybe_model, CustomObjectiveModel):
            return maybe_model
        else:
            return CustomObjectiveFunction(func, maybe_model)


"""
    TempInspection.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
from time import sleep
from DataUtils import get_in_folder
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry

fh = None

def datarange_problem_inspection(agent, result_folder):
    global fh

    in_folder = get_in_folder(agent.in_folder.get())
    print("\n\n\n------------------- datarange_problem_inspection", in_folder)
    print("result_folder=", result_folder)

    agent.show_datarange_dialog(__wait__=False)

    while not agent.has_datarange_dialog():
        print("waiting for the datarange_dialog being ready")
        sleep(1)

    dialog = agent.datarange_dialog

    xr_frame = dialog.get_current_frame()
    xr_frame.toggle_btn.invoke()
    sleep(1)

    uv_frame = dialog.get_current_frame()
    # canvases[0]: class_3d
    # canvases[1]: class_2d - upper right frame
    # canvases[2]: class_2d - lower right frame
    uv_frame.canvases[1].from_opposite_side_btn.invoke()

    dialog.cancel()
    print("------------------- canceled")
    sleep(1)

    if fh is None:
        # clear_dirs_with_retr([result_folder])
        path = os.path.join(result_folder, "inspected.csv")
        fh = open(path, "w")
        fh.write("folder\n")
    fh.write(",".join([in_folder]) + "\n")

    agent.analysis_button.invoke(__wait__=False)
    sleep(1)
    analyzer = agent.analyzer

    while not analyzer.has_mapper_canvas():
        sleep( 1 )

    analyzer.mapper_canvas.cancel()
    sleep( 1 )

def inspection_close(agent):
    fh.close()
    print("------------------- closed\n\n\n")



# coding: utf-8
"""
    ConjugateGradientThreeD.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
import numpy as np
try:
    # for numba 1.49 or later
    from numba.core.decorators import jit
except:
    from numba.decorators import jit
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import matplotlib.patches as mpl_patches
import matplotlib.animation as animation
from molass_legacy.KekLib.OurTkinter import Tk, Dialog

TITLE_FONTSIZE  = 16
USE_3D_WORKAROUND   = False

def  dot_plot_wa(ax, x, y, z, marker=None, color=None, markersize=None):
    ax.plot([x[0], x[0]], [y[0], y[0]], [z[0], z[0]], marker=marker, color=color, markersize=markersize)

@jit(nopython=True)
def compute_norm_1d(x, y, elution_z, mx, my):
    assert x.shape == y.shape

    z = []
    for i in range(x.shape[0]):
        x_ = x[i]
        y_ = y[i]
        z_ = np.linalg.norm( elution_z * x_ - mx )**2 + np.linalg.norm( elution_z * y_ - my )**2 
        z.append(z_)

        """
        if False and i == x.shape[0]//2:
            fig = dplt.figure()
            ax = fig.gca()
            ax.plot(elution_z * x_, label='elution_z * x_')
            ax.plot(mx, label='mx')
            ax.plot(elution_z * y_, label='elution_z * y_')
            ax.plot(my, label='my')
            ax.legend()
            fig.tight_layout()
            dplt.show()
        """

    return np.array(z)

@jit(nopython=True)
def compute_norm_2d(aa, bb, elution_z, ma, mb):
    assert aa.shape == bb.shape

    ret_array = np.zeros(aa.shape)
    # row_array = []
    for i in range(aa.shape[0]):
        # row = []
        for j in range(aa.shape[1]):
            a = aa[i,j]
            b = bb[i,j]
            c = np.linalg.norm( elution_z * a - ma )**2 + np.linalg.norm( elution_z * b - mb )**2 
            """
            if False and i == aa.shape[0]//2 and j == aa.shape[1]//2:
                fig = dplt.figure()
                ax = fig.gca()
                ax.plot(elution_z * a, label='elution_z * a')
                ax.plot(ma, label='ma')
                ax.plot(elution_z * b, label='elution_z * b')
                ax.plot(mb, label='mb')
                ax.legend()
                fig.tight_layout()
                dplt.show()
            """
            # row.append(c)
            ret_array[i,j] = c
        # row_array.append(row)

    # return np.array(row_array)
    return ret_array

dp = 0.03

class GdTreedPlot:
    def __init__(self, dialog, draw_trace):
        self.q  = dialog.q
        self.qf = dialog.qf
        self.qt = dialog.qt
        self.p1 = dialog.p1
        self.p2 = dialog.p2
        self.eval_range = dialog.eval_range
        self.data = dialog.data
        self.elution_z = dialog.elution_z
        self.p2m_scale = dialog.p2m_scale
        self.draw_trace = draw_trace

        N = 21
        p = np.linspace(self.p1 - dp, self.p1 + dp, N)
        q = np.linspace(self.p2 - dp, self.p2 + dp, N)
        pp, qq = np.meshgrid(p, q)
        f, _, t = self.eval_range
        f_data = self.data[self.qf, f:t]
        t_data = self.data[self.qt, f:t]
        rr = compute_norm_2d(pp, qq, self.elution_z, f_data, t_data)
        self.surface_xyz = (pp, qq, rr)

        if draw_trace:
            points = np.array(dialog.solver.demo_data)
            print('points.shape=', points.shape)
            x = points[:,0]*self.p2m_scale
            y = points[:,1]*self.p2m_scale
            z = compute_norm_1d(x, y, self.elution_z, f_data, t_data)
            self.trace_xyz = (x, y, z)

    def draw(self, ax, title_fontsize, animate_fig=None, title_add=''):
        qf_ = self.q[self.qf]
        qt_ = self.q[self.qt]

        ax.set_title("Objective function restricted to A(%.2g) and A(%.2g)%s" % (qf_, qt_, title_add), fontsize=title_fontsize)
        ax.set_xlabel('$A(%.2g)$' % qf_)
        ax.set_ylabel('$A(%.2g)$' % qt_)
        ax.set_zlabel('$norm^2$')

        ax.plot([self.p1 - dp, self.p1 + dp], [self.p2 - dp, self.p2 - dp], [0, 0], ':', color='red', alpha=0.3, linewidth=5)
        if USE_3D_WORKAROUND:
            dot_plot_wa(ax, [self.p1], [self.p2 - dp], [0], 'o', color='red', markersize=10)
        else:
            ax.plot([self.p1], [self.p2 - dp], [0], 'o', color='red', markersize=10)

        ax.plot([self.p1 + dp, self.p1 + dp], [self.p2 - dp, self.p2 + dp], [0, 0], ':', color='green', alpha=0.3, linewidth=5)
        if USE_3D_WORKAROUND:
            dot_plot_wa(ax, [self.p1 + dp], [self.p2], [0], 'o', color='green', markersize=10)
        else:
            ax.plot([self.p1 + dp], [self.p2], [0], 'o', color='green', markersize=10)

        ax.plot_surface(*self.surface_xyz, color='pink', alpha=0.3)

        if self.draw_trace:
            if animate_fig is None:
                ax.plot(*self.trace_xyz, marker='o', color='red', markersize=3)
            else:
                self.animate_trace_3d(animate_fig, ax)

    def animate_trace_3d(self, fig, ax, add_inset=True):
        x, y, z = self.trace_xyz
        ax.plot(x, y, z, marker='o', color='yellow', markersize=3)
        line, = ax.plot([x[0], x[0]],[y[0], y[0]], [z[0], z[0]], marker='o', color='red', markersize=3)

        data = np.array([x, y, z])

        def update_line(n, data, line):
            line.set_data(data[0:2, :n])
            line.set_3d_properties(data[2, :n])
            return (line,)

        self.anim3d = animation.FuncAnimation(fig, update_line, len(x), fargs=(data, line),
                    interval=200, blit=True)

        if add_inset:
            self.animate_trace_2d(fig, ax)

    def animate_trace_2d(self, fig, ax3d):
        ax = ax3d.inset_axes([0.04, 0.65, 0.25, 0.25])
        ax.set_xticklabels('')
        ax.set_yticklabels('')

        x, y, z = self.trace_xyz

        last_x = x[-1]
        last_y = y[-1]
        w, h = 0.0002, 0.0002
        ax.set_xlim(last_x - w, last_x + w)
        ax.set_ylim(last_y - h, last_y + h)

        ax.plot(x, y, marker='o', color='yellow', markersize=3)
        line, = ax.plot([x[0], x[0]], [y[0], y[0]], marker='o', color='red', markersize=5)

        data = np.array([x, y])

        def update_line(n, data, line):
            line.set_data(data[:, :n])
            return (line,)

        self.anim2d = animation.FuncAnimation(fig, update_line, len(x), fargs=(data, line),
                    interval=200, blit=True)

class GdAnimationDialog(Dialog):
    def __init__(self, parent, gd3d):
        self.parent = parent
        self.gd3d   = gd3d

        Dialog.__init__( self, parent, "Gradient Descent Animation", visible=False )

    def __del__(self):
        """
        TODO:
            seems to be a bug in tkinter.__init__.py
            fix tkinter instead of fixing app

        Exception in Tkinter callback
        Traceback (most recent call last):
          File "C:\Program Files\Python37\lib\tkinter\__init__.py", line 1705, in __call__
            return self.func(*args)
          File "C:\Program Files\Python37\lib\tkinter\__init__.py", line 752, in callit
            self.deletecommand(name)
          File "C:\Program Files\Python37\lib\tkinter\__init__.py", line 601, in deletecommand
            self._tclCommands.remove(name)
        AttributeError: 'NoneType' object has no attribute 'remove'
        """
        pass

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(12, 10))

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.mpl_canvas.draw()
        ax = fig.add_subplot(111, projection='3d')
        self.gd3d.draw(ax, TITLE_FONTSIZE, animate_fig=fig, title_add=' with a zoomed 2D inset')

        ax.disable_mouse_rotation()     # to avoid a rotation bug

        fig.tight_layout()

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.bind("<Return>", self.ok)



"""
    SecTheory.BasicProofs.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import iv
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt

def simple_pdf(t, npi, tpi):
    return iv(1, np.sqrt(4*npi*t/tpi)) * np.sqrt(npi/(t*tpi)) * np.exp(-t/tpi-npi)

def numerical_inversion_proof(**kwargs):
    from .SecCF import simple_phi
    from .SecPDF import FftInvPdf

    fig, ax = plt.subplots()

    ax.set_title("Numerical Inversion Proof (1)", fontsize=16)

    x = np.arange(10, 300)
    npi = 20
    tpi = 5

    pdf = FftInvPdf(simple_phi)

    y1 = simple_pdf(x, npi, tpi)
    ax.plot(x, y1, label="formula pdf(t)")
    y2 = pdf(x, npi, tpi)
    ax.plot(x, y2, label="numerically inverted pdf(t)")
    ax.legend()

    print("sum(y1)=", np.sum(y1))
    print("sum(y2)=", np.sum(y2))

    fig.tight_layout()
    plt.show()

def roud_trip(in_folder=None):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder
    from SecSaxs.DataTreatment import DataTreatment
    from .SecPDF import compute_standard_wCD

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=2, correction=1)
    sd_copy = treat.get_treated_sd(sd, pre_recog)

    D, E, qv, ecurve = sd_copy.get_xr_data_separate_ly()

    fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(20,5))
    fig.suptitle("Round Trip Proof for %s" % get_in_folder(in_folder), fontsize=20)

    ax1.set_title("Elution", fontsize=16)
    ax2.set_title("PDF", fontsize=16)
    ax3.set_title("CF(real)", fontsize=16)
    ax4.set_title("CF(imag)", fontsize=16)

    x = ecurve.x
    y = ecurve.y
    ax1.plot(x, y, label="data")

    area = np.sum(y)
    y_ = y/area
    ax2.plot(x, y_, label="pdf")

    N = 1024
    w, C, D = compute_standard_wCD(N)
    cft = []
    for w_ in w:
        cft.append(np.sum(np.exp(1j*w_*x)*y_))

    cft = np.array(cft)
    pdfFFT = np.max([np.zeros(N), (C*np.fft.fft(D*cft)).real], axis=0)
    spline = UnivariateSpline(np.arange(N), pdfFFT, s=0)
    inv_y = spline(x)
    ax2.plot(x, inv_y, label="inverted pdf")

    ax1.plot(x, inv_y*area, label="round tripped")

    ax3.plot(w, cft.real)
    ax4.plot(w, cft.imag)

    for ax in [ax1, ax2]:
        ax.legend()

    fig.tight_layout()
    plt.show()



"""

    LambertBeerTester.py

    Copyright (c) 2024, SAXS Team, KEK-PF

"""
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt

def test_compute_base_plane(data, index, ecurve, denoise=False):
    BP = None
    def test_compute_base_plane_impl():
        import Baseline.LambertBeer
        reload(Baseline.LambertBeer)
        from molass_legacy.Baseline.LambertBeer import BasePlane
        nonlocal BP
        bp = BasePlane(data, index, ecurve, denoise=denoise, debug=True)
        bp.solve(debug=True)
        BP = bp.get_baseplane()
        return BP

    with plt.Dp(extra_button_specs=[("Test", test_compute_base_plane_impl)]):
        fig, ax = plt.subplots()
        ax.set_title("test_compute_base_plane")
        plt.show()

    return BP


"""
    SecTheory.SecParamsSurvey.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np

columntype_id = None

def inspect_sec_params(in_folder, *args, **kwargv):
    from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings, clear_v2_temporary_settings, set_setting, get_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Batch.LiteBatch import LiteBatch
    from molass_legacy.SerialAnalyzer.DataUtils import cut_upper_folders
    from molass_legacy.Tools.EmbedDialog import open_dialog
    from Experiment.ColumnTypes import update_sec_settings_by_id

    global fh, count, columntype_id

    count += 1
    print([count], in_folder)
    set_setting('in_folder', in_folder)

    if count < start:
        return True

    folder_name = cut_upper_folders(in_folder)
    # if False:
    # if folder_name < "20170506":
    # if folder_name.find("OA001") < 0:

    if True:
        if folder_name.find("proteins5") >= 0:
            columntype_id = "ad6w"
        else:
            print("skip")
            return True

    try:
        clear_temporary_settings()
        clear_v2_temporary_settings()

        if columntype_id is None:
            columntype_id = "ad200w"
        update_sec_settings_by_id(columntype_id)
        poresize = get_setting('poresize')
        print("poresize (1)", poresize)

        sp = StandardProcedure()
        sd = sp.load_old_way(in_folder)

        lb = LiteBatch()
        lb.prepare(sd, num_peaks=6)

        init_params = lb.get_init_estimate()

        optimizer = lb.get_optimizer()
        optimizer.prepare_for_optimization(init_params, minimally=True)

        print("init_separate_params[7]", optimizer.init_separate_params[7])

        lrf_info = lb.optimizer.objective_func(init_params, return_lrf_info=True)

        # plot_sec_params(optimizer, init_params, lrf_info)

        def reload_wrapper():
            from importlib import reload
            import SecTheory.SecParamsPlot
            reload(SecTheory.SecParamsPlot)
            from SecTheory.SecParamsPlot import plot_sec_params
            plot_sec_params(optimizer, init_params, lrf_info)

        open_dialog(reload_wrapper)

    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "inspect_sec_params: ", n=10)

    fh.flush()

    return False

def collect_sec_params(input_folder, result_folder):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from DevelUtils.FolderWalkers import serial_folder_walk
    global fh, count, start

    result_file = os.path.join(result_folder, "sec-params.csv")

    fh = open(result_file, "w")
    count = -4
    start = 0

    serial_folder_walk(input_folder, inspect_sec_params, reverse=True)

    fh.close()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()

    import molass_legacy.KekLib, SerialAnalyzer, DataStructure, GuinierAnalyzer, Decomposer, Extrapolation
    from molass_legacy.SerialAnalyzer.DataUtils import get_pytools_folder

    root_folder = get_pytools_folder()
    input_folder = os.path.join(root_folder, "Data")
    for drive in ['D', 'E']:
        result_folder = r"%s:\TODO\20231218\sec-params" % drive
        if os.path.exists(result_folder):
            break
    collect_sec_params(input_folder, result_folder)



import numpy                as np
from mpl_toolkits.mplot3d   import Axes3D
import matplotlib.pyplot    as plt
from Intensity              import Intensity

Q = np.arange( 0, 0.8, 1e-3 )
Q_slice = slice( 0, 100 )
C_array = [ 1e-2, 1e-3 ]

X = Q**2

fig = plt.figure( figsize=(24,12) )
ax0 = fig.add_subplot( 231 )
ax1 = fig.add_subplot( 232,  projection='3d' )
ax2 = fig.add_subplot( 233,  projection='3d' )

axes = []
axes.append( [ ax0, ax1, ax2 ] )

ax0 = fig.add_subplot( 234 )
ax1 = fig.add_subplot( 235,  projection='3d' )
ax2 = fig.add_subplot( 236,  projection='3d' )
axes.append( [ ax0, ax1, ax2 ] )

for i, C in enumerate(C_array):
    conc = 0.2 * np.exp( - C * ( np.arange( 100 ) - 50 )**2  )

    Y1_array = []
    Y__array = []

    for c in conc:

        intensity = Intensity( c, 29 )

        y1 = intensity.term1( Q )
        y2 = intensity.term2( Q )

        ii = y1 * y2

        Y1 = np.log( y1 )
        Y_ = np.log( y1 - ii )

        Y1_array.append( Y1 )
        Y__array.append( Y_ )

    ax0, ax1, ax2 = axes[i]
    ax0.set_title( 'Concentration Variation' )
    ax0.set_xlabel( 'Seq-No.' )
    ax0.set_ylabel( 'Concentration' )
    ax1.set_title( 'Guinier plot varying Seq-No.' )
    ax1.set_xlabel( 'Q' )
    ax1.set_ylabel( 'Seq-No.' )
    ax1.set_zlabel( 'Intensity' )
    ax2.set_title( 'Guinier plot varying Concentration' )
    ax2.set_xlabel( 'Q' )
    ax2.set_ylabel( 'Concentration' )
    ax2.set_zlabel( 'Intensity' )

    ax0.plot( conc, color='blue' )

    for i in range( 100 ):
        X_ = X[Q_slice]
        Y1 = Y1_array[i][Q_slice]
        Y_ = Y__array[i][Q_slice]
        I_ = np.ones( ( X_.shape[0], ) ) * i
        C  = np.ones( ( X_.shape[0], ) ) * conc[i]
        ax1.plot( X_, I_, Y_, color='red' )
        ax1.plot( X_, I_, Y1, ':', color='blue' )
        ax2.plot( X_, C, Y_, color='red' )
        ax2.plot( X_, C, Y1, ':', color='blue' )

plt.tight_layout()
plt.show()



"""
    ProgressMinDialog.py

    Copyright (c) 2018-2024, Masatsuyo Takahashi, KEK-PF
"""
import sys
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkSupplements import set_icon

class ProgressMinDialog(Dialog):
    def __init__(self, parent, title="ProgressMinDialog", message=None, num_steps=10, length=200,
                    progress_cb=None, geometry_info=None, cancelable=False, interval=500,
                    visible=True):
        self.parent = parent
        self.num_steps = num_steps
        self.length = length
        self.canceled = False
        self.message = message
        self.progress_cb = progress_cb
        self.geometry_info = geometry_info
        self.cancelable = cancelable
        self.interval = interval
        auto_geometry = geometry_info is None
        Dialog.__init__( self, self.parent, title, auto_geometry=auto_geometry, visible=visible)

    def show(self):
        """
        used when self has been constructed with visible=False
        """
        self._show()

    def body(self, body_frame):   # overrides Dialog.body
        set_icon(self)
        """
            https://stackoverflow.com/questions/7310511/how-to-create-downloading-progress-bar-in-ttk
        """
        if self.message is not None:
            label = Tk.Label(self, text=self.message)
            label.pack()

        self.mpb = ttk.Progressbar(body_frame,orient ="horizontal", length=self.length, mode="determinate")
        self.mpb.pack(padx=5, pady=5)
        self.mpb["maximum"] = self.num_steps
        self.update()

        if self.geometry_info is not None:
            self.geometry(self.geometry_info)
            self.update()

        self.step = 0
        self.refresh()

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        if self.cancelable:
            w = Tk.Button(box, text="Cancel", width=10, command=lambda: self.cancel(ask=True) )
            w.pack(side=Tk.LEFT, padx=5, pady=5)

    def cancel(self, ask=False):
        if ask:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            yn = MessageBox.askyesno( "Cancel comfirmation",
                "Are you sure to cancel this process?", parent=self )
            if yn:
                pass
            else:
                return

        self.canceled = True
        super(ProgressMinDialog, self).cancel()

    def refresh(self):
        self.mpb["value"] = self.step
        if self.progress_cb is None:
            self.step += 1
        else:
            self.step = self.progress_cb()

        if self.step < 0:
            self.cancel()
        elif self.step >= self.num_steps:
            self.ok()
        else:
            self.after( self.interval, self.refresh )
            self.update()

def run_with_progress(parent, proc, max_iter=10, title=None, on_return=None, debug=True):
    """
    TODO:
        make this cancelable for cases where it takes a long time.
    """
    import queue
    import threading
    from molass_legacy.KekLib.TkUtils import split_geometry
    if debug:
        import logging
        logger = logging.getLogger(__name__)

    w, h, x, y = split_geometry( parent.winfo_geometry() )
    geometry_info = "+%d+%d" % (parent.winfo_rootx() + int(w*0.7), parent.winfo_rooty() + int(h*0.8))

    exe_queue = queue.Queue()

    counter = [0]
    ret_info = [None, None]
    def progress_cb():
        # print('progress_cb:')
        error_info = None
        try:
            info = exe_queue.get(block=False)
            if debug:
                logger.info('progress_cb:', info[0])
            if info[0] == 1:
                ret_info[0] = info[1]
                if on_return is not None:
                    on_return(ret_info[0])
                counter[0] = max_iter + 1
            elif info[0] == 0:
                counter[0] = info[1] + 1
            else:
                # close the progressbar
                counter[0] = -1
                ret_info[1] = info[1]
        except:
            pass
        return counter[0]

    def thread_proc():
        try:
            ret = proc(exe_queue)
            exe_queue.put([1, ret])
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            exe_queue.put([-1, etb])

    thread = threading.Thread( target=thread_proc, name='ThreadInProgress', args=[] )
    thread.start()

    progress = ProgressMinDialog(parent,
                    title=title,
                    num_steps=max_iter + 1,
                    progress_cb=progress_cb, geometry_info=geometry_info,
                    visible=False )

    progress.show()

    thread.join()

    ret_tuple, error_info = ret_info
    return ret_tuple, error_info



"""

    OutlineFigureSaver.py

    Copyright (c) 2022, SAXS Team, KEK-PF

"""
import numpy as np
import imageio.v3 as iio
import molass_legacy.KekLib.DebugPlot as plt

def save_the_figure_trial_impl( self, folder, analysis_name ):
    # print( 'save_the_figure: ', folder, analysis_name )
    filename = analysis_name.replace( 'analysis', 'figure' )
    path = os.path.join( folder, filename )

    """
    Matplotlib figure to image as a numpy array
    https://stackoverflow.com/questions/35355930/matplotlib-figure-to-image-as-a-numpy-array
    """
    s, (width, height) = self.mpl_canvas.print_to_buffer()

    # Option 2a: Convert to a NumPy array.
    image = np.fromstring(s, np.uint8).reshape((height, width, 4))
    print('image.shape=', image.shape)

    """
    How to Resize, Pad Image to Square Shape and Keep Its Aspect Ratio with Python
    https://jdhao.github.io/2017/11/06/resize-image-to-square-with-padding/
    """
    import cv2
    white = [255, 255, 255]
    top, bottom, left, right = 50, 10, 10, 10
    new_im = cv2.copyMakeBorder(image, top, bottom, left, right, cv2.BORDER_CONSTANT,
        value=white)

    if True:
        import molass_legacy.KekLib.DebugPlot as plt
        plt.show()
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.imshow(new_im)
            plt.show()
    else:
        cv2.imshow("image", new_im)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    cv2.imwrite(path + '.png', new_im)

def save_the_upper_figure_impl(self):
    from tkinter import filedialog
    from DataUtils import get_in_folder
    """
    Matplotlib figure to image as a numpy array
    https://stackoverflow.com/questions/35355930/matplotlib-figure-to-image-as-a-numpy-array
    """
    buf = self.mpl_canvas.buffer_rgba()

    # convert to a NumPy array
    outline_image = np.asarray(buf)
    h, w = outline_image.shape[0:2]

    in_folder = get_in_folder()

    # using DebugPlot as a tool for internal drawing
    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title(in_folder, fontsize=16)
        ax.set_axis_off()
        ax.imshow(outline_image)
        fig.tight_layout()
        fig.canvas.draw()
        buf = fig.canvas.buffer_rgba()
        X = np.asarray(buf)
        H, W = X.shape[0:2]
        hc = H//2 
        wc = W//2
        hslice = slice(0, hc+30)
        sw = w//2 + 30
        wslice = slice(wc-sw, wc+sw)
        upper_image = X[hslice,wslice,:]
        upper_image[-20:,:,:] = 255
        # plt.show()

    files = [('PNG Files', '*.png'), ('JPEG Files', '*.jpg')]
    filename = in_folder.split("/")[-1] + ".png"
    path = filedialog.asksaveasfilename(parent=self.parent, initialfile =filename, filetypes=files, defaultextension=files)
    iio.imwrite(path, upper_image)

def save_the_elution_curve_impl(self, type):
    from tkinter import filedialog
    from DataUtils import get_in_folder

    sd = self.serial_data
    if type == 0:
        ecurve = sd.get_xray_curve()
    else:
        ecurve = sd.get_uv_curve()

    x = ecurve.x
    y = ecurve.y

    in_folder = get_in_folder()

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title(in_folder, fontsize=16)
        ax.plot(x, y)
        fig.tight_layout()
        plt.show()



"""
    Selective/RdrMinimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from Selective.V1ParamsAdapter import make_decomp_result_impl

def try_rdr_minimization(advanced_frame, devel=True):
    if devel:
        from importlib import reload
        import Selective.MinimizerUtils
        reload(Selective.MinimizerUtils)
    from .MinimizerUtils import RgComputer

    print("try_rdr_minimization")
    editor = advanced_frame.editor
    editor_frame = editor.get_current_frame()
    model = editor_frame.model
    print("edm_text_from_editor_frame: ", model.get_name(), model.__class__)
    params_array = editor.get_current_params_array()

    fx = editor_frame.fx
    x = editor_frame.x
    y = editor_frame.y
    uv_y = editor_frame.uv_y

    props = model.get_proportions(fx, params_array)
    i = np.argmax(props)    # target component
    print("props=", props, "i=", i)

    paired_ranges = editor_frame.make_restorable_ranges()
    print("paired_ranges=", paired_ranges)
    compute_range = paired_ranges[i]

    D, E, qv, ecurve = editor.sd.get_xr_data_separate_ly()
    num_components = len(params_array)
    rg_computer = RgComputer(model, D, E, qv, fx, y, compute_range, num_components, i)
    C = rg_computer.compute_C(params_array)

    def plot_params(params_array, title, **dp_wkargs):
        rg_list, sg_list = rg_computer.compute_rg_list(C, return_sg=True)
        print("rg_list=", rg_list)
        rdr = abs(rg_list[0] - rg_list[1])*2/(rg_list[0] + rg_list[1])
        with plt.Dp(**dp_wkargs):
            fig, ax = plt.subplots()
            ax.set_title(title, fontsize=20)
            ax.plot(fx, y)
            areas = []
            for cy in C:
                 areas.append(np.sum(cy))
            props = np.array(areas)/np.sum(areas)
            for k, cy in enumerate(C):
                 ax.plot(fx, cy, ":", label="component-%d (%.2g)" % (k, props[k]))
            ty = np.sum(C, axis=0)
            ax.plot(fx, ty, ":", color="red", label="model total")
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = xmin*0.9 + xmax*0.1
            ty = ymin*0.3 + ymax*0.7
            ax.text(tx, ty, "RDR=%.2g" % rdr, fontsize=16)
            ax.legend(fontsize=16)
            fig.tight_layout()
            ret = plt.show()
        return ret

    advanced_frame.progress_update(1)

    plot_params(params_array, "try_rdr_minimization entry")

    rg_list, sg_list = rg_computer.compute_rg_list(C, return_sg=True)
    print("before rg_list=", rg_list)
    sg = sg_list[i]
    gslice = slice(sg.guinier_start, sg.guinier_stop)

    def objective(p, debug=False):
        params_array_ = p.reshape(params_array.shape)
        C_ = rg_computer.compute_C(params_array_)
        ty = np.sum(C_, axis=0)
        try:
            rg_list = rg_computer.compute_rg_list(C_, gslice=gslice)
            rg_ = (rg_list[0] + rg_list[1])/2
        except:
            # numpy.linalg.LinAlgError: SVD did not converge 
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "try_rdr_minimization.objective: ")
            rg_ = 0
        if rg_ > 10:
            rdr = abs(rg_list[0] - rg_list[1])/rg_
            # scale, N, T, Rg, x0
            # param_rg = params_array_[i,3]
            # fv = np.mean((ty - y)**2) * rdr
            # return fv * max(fv, (rg_ - param_rg)**2)
            ydiv = np.log10(np.mean((ty - y)**2))
            rdiv = np.log10(rdr)
            fv = ydiv*0.95 + rdiv*0.05
            if debug:
                print("rdr=", rdr)
                print("ydiv, rdiv=", ydiv, rdiv, fv)
            return fv
        else:
            fv = np.inf
        return fv

    init_params_array = params_array.flatten()
    objective(init_params_array, debug=True)
    advanced_frame.progress_update(2)

    call_count = 0
    def minima_callback(x, f, accept):
        nonlocal call_count
        call_count += 1
        advanced_frame.progress_update(call_count + 2)

    minimizer_kwargs = dict(method='Nelder-Mead')
    ret = basinhopping(objective, init_params_array, niter=10, callback=minima_callback,
                       minimizer_kwargs=minimizer_kwargs)
    advanced_frame.progress_update(14)

    opt_params_array = ret.x.reshape(params_array.shape)
    objective(opt_params_array, debug=True)
    optC = rg_computer.compute_C(opt_params_array)
    rg_list = rg_computer.compute_rg_list(optC, gslice=gslice)
    print("after rg_list=", rg_list)
    advanced_frame.progress_update(15)
    dp_kwargs = dict(window_title="RDR minimization", 
                     button_spec=["Accept", "Cancel"], 
                     guide_message="If you accept, the model will be updated with the new parameters.")
    ret = plot_params(opt_params_array, "Minimized RDR on %s" % get_in_folder(), **dp_kwargs)
    if ret:
        advanced_frame.progress_update(16)
        decomp_result = make_decomp_result_impl(editor, opt_params_array)
        advanced_frame.progress_update(17)
        advanced_frame.update_button_status(change_id="RDR")
        editor.update_current_frame_with_result(decomp_result)
        advanced_frame.progress_update(18)
    else:
        advanced_frame.progress_update(18)
    advanced_frame.progress_update(-1)


"""
    Optimizer.FunctionDebugger.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Slider
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from SecTheory.BasicModels import robust_single_pore_pdf as monopore_pdf
from molass_legacy.Baseline.BaselineUtils import create_xr_baseline_object
from molass_legacy._MOLASS.SerialSettings import get_setting

class FunctionDebugger(Dialog):
    def __init__(self, parent, dialog, js_canvas, composite=None, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.FunctionDebuggerUtils
            reload(Optimizer.FunctionDebuggerUtils)
        from .FunctionDebuggerUtils import create_optimizer_for_debug

        self.parent = parent
        self.dialog = dialog
        self.js_canvas = js_canvas
        self.dsets = dsets = js_canvas.dsets
        ecurve = dsets[0][0]
        self.x = ecurve.x
        self.y = ecurve.y
        self.elution_model = js_canvas.elution_model

        self.fullopt = create_optimizer_for_debug(
                            js_canvas.dsets, js_canvas.fullopt.n_components,
                            js_canvas.optinit_info, js_canvas.optinit_info.init_params,
                            composite=composite
                            )

        self.params = js_canvas.get_current_params()
        self.separate_params = self.fullopt.split_params_simple(self.params)
        Dialog.__init__(self, parent, "Function Debugger", visible=False)

    def get_xr_params(self):
        return self.separate_params[0]

    def get_rg_params(self):
        return self.separate_params[2]

    def get_sec_params(self):
        return self.separate_params[-1]

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        fig, axes = plt.subplots(ncols=3, figsize=(18,5))
        self.fig = fig

        axt = axes[1].twinx()
        axt.grid(False)
        self.axes = list(axes)
        self.axes.append(axt)

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.popup_menu = None
        self.start_debugger()

    def start_debugger(self):
        from importlib import reload

        import Optimizer.TheDebugUtils
        reload(Optimizer.TheDebugUtils)
        import Optimizer.FvScoreConverter
        reload(Optimizer.FvScoreConverter)

        from .FvScoreConverter import convert_score

        if self.elution_model in [0, 2]:
            import ModelParams.EghPlotUtils as plot_utils
        else:
            import ModelParams.StcPlotUtils as plot_utils
        reload(plot_utils)

        if False:
            import Optimizer.CompositeInfo
            reload(Optimizer.CompositeInfo)
            from .CompositeInfo import CompositeInfo

            composites = [[0], [1,2], [3], [4]]
            composite = CompositeInfo(composites=composites)
            optworking_folder = get_setting('optworking_folder')
            path = os.path.join(optworking_folder, 'composite_info.txt')
            composite.save(path)
            composite.load(path)
            assert str(composites) == str(composite.composites)

        # should be unified with JobStateCanvas.draw_indexed_state()
        fig = self.fig
        axes = self.axes
        ax1, ax2, ax3, axt = self.axes
        ax1.set_title("UV Decomposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)
        fv = self.fullopt.objective_func(self.params)
        ax3.set_title("Objective Function Scores in SV=%.3g" % convert_score(fv), fontsize=16)

        self.temp_params = self.params.copy()
        self.temp_params[4*3+2] = 35.9
        self.fullopt.objective_func(self.temp_params, plot=True, axis_info=(self.fig, self.axes))

        fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, padx=20)

        num_buttons = 4

        for j in range(num_buttons):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        col = 0
        w.grid(row=0, column=col, pady=10)
        self.cancel_btn = w

        col += 1
        w = Tk.Button(box, text="Show Bounds Inspection", width=24, command=self.show_bounds_indpection)
        w.grid(row=0, column=col, pady=10)

        col += 1
        w = Tk.Button(box, text="Show Normalized Parameters", width=24, command=self.show_norm_pameters)
        w.grid(row=0, column=col, pady=10)

        col += 1
        w = Tk.Button(box, text="Show Parameters", width=24, command=self.show_parameters)
        w.grid(row=0, column=col, pady=10)

        col += 1
        w = Tk.Button(box, text="Test", width=10, command=self.test)
        w.grid(row=0, column=col, pady=10)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show_bounds_indpection(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.BoundsInspection
            reload(Optimizer.BoundsInspection)
        from molass_legacy.Optimizer.BoundsInspection import BoundsInspection
        parent = self.js_canvas.dialog.parent
        dialog = BoundsInspection(parent, self.fullopt, self.temp_params)
        dialog.show()

    def show_norm_pameters(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.ParamsInspection
            reload(Optimizer.ParamsInspection)
        from .ParamsInspection import ParamsInspection
        # self.dialog.grab_set()  # temporary fix to the grab_release problem
        norm_params = self.fullopt.to_norm_params(self.temp_params)
        parent = self.js_canvas.dialog.parent
        dialog = ParamsInspection(parent, norm_params, self.dsets, self.fullopt)     # consider better positioning (or positioning bug?)
        dialog.show()
        # self.dialog.grab_set()  # temporary fix to the grab_release problem

    def show_parameters(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.ParamsInspection
            reload(Optimizer.ParamsInspection)
        from .ParamsInspection import ParamsInspection
        # self.dialog.grab_set()  # temporary fix to the grab_release problem
        parent = self.js_canvas.dialog.parent
        dialog = ParamsInspection(parent, self.temp_params, self.dsets, self.fullopt)     # consider better positioning (or positioning bug?)
        dialog.show()
        # self.dialog.grab_set()  # temporary fix to the grab_release problem

    def test(self):
        from importlib import reload
        import Optimizer.StructureFactorBoundsDemo
        reload(Optimizer.StructureFactorBoundsDemo)
        from .StructureFactorBoundsDemo import demo, demo_for_debugger

        self.dialog.grab_set()  # temporary fix to the grab_release problem
        demo_for_debugger(self.dialog, self.js_canvas)
        self.dialog.grab_set()  # temporary fix to the grab_release problem



# coding: utf-8
"""
    DatasetFilter.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
import re
from openpyxl import load_workbook
from DataUtils import cut_upper_folders

class DatasetFilter:
    def __init__(self, book_path):
        wb = load_workbook(book_path)
        ws = wb.active
        dataset_dict = {}
        for i in range(2, 500):
            row = []
            for j in [1, 2, 3]:
                c = ws.cell(row=i, column=j)
                if c.value is None:
                    break
                row.append(c.value)
            if len(row) == 0:
                break

            if row[2] != 'O':
                continue

            key = row[0][0:8]
            count = dataset_dict.get(key, 0)
            dataset_dict[key] = count+1

        self.dataset_dict = dataset_dict
        self.find_extra_info(os.path.dirname(book_path))

    def find_extra_info(self, folder):
        filter_txt = os.path.join(folder, "filter.txt")
        self.extra_dict = None
        if not os.path.exists(filter_txt):
            return

        in_folder_re = re.compile(r'(\S+)')
        self.extra_dict = {}
        fh = open(filter_txt)
        for line in fh:
            print(line)
            m = in_folder_re.search(line)
            if m:
                in_folder = m.group(1)
                nodes = in_folder.split('/')
                self.extra_dict[nodes[0][0:8]] = 1
        fh.close()

    def is_in_the_selection(self, in_folder):
        in_folder_ = cut_upper_folders(in_folder)
        nodes = in_folder_.split('/')
        key = nodes[0][0:8]
        count = self.dataset_dict.get(key, 0)
        if self.extra_dict is None:
            ret = count > 0
        else:
            ret = count > 0 and self.extra_dict.get(key, 0) == 1
        return ret



# coding: utf-8
"""
    AnalysisRangeManager.py

    Copyright (c) 2016-2020, SAXS Team, KEK-PF
"""
import numpy                as np
from molass_legacy.SerialAnalyzer.ElutionCurve           import ElutionCurve
from molass_legacy.DataStructure.AnalysisRangeInfo      import convert_to_paired_ranges
from PairedRangeLogger      import log_paired_ranges

DEBUG   = False

def get_analysis_ranges_from_sd( self, range_type, min_value, quality_array=None, conc_y=None, rg_array=None, ic_array=None ):
    # print( 'get_analysis_ranges_from_sd: min_quality=', min_quality )

    if conc_y is None:
        conc_y = self.conc_curve.y

    if range_type == 0:
        # ref_curve specification excludes invalid subobjects
        # which have inappropriate ref_curve values.
        curve = ElutionCurve( quality_array, low_quality=True )
    elif range_type == 1:
        curve = ElutionCurve( conc_y )
    else:
        assert( False )

    # print( 'curve.subobject_array=', curve.subobject_array )
    ranges = []
    for i, subobj in enumerate( curve.subobject_array ):
        ft, obj = subobj
        # print( 'subobj[%d]' % i, ft, obj.lower_abs, obj.middle_abs, obj.upper_abs )
        ranges.append( [ obj.lower_abs, obj.middle_abs, obj.upper_abs ] )

    return ranges

def get_analysis_ranges_for_exec( self, range_type, analysis_ranges, logger ):
    assert analysis_ranges is not None

    ranges = analysis_ranges
    ret = convert_to_paired_ranges( ranges )
    paired_ranges = ret[0]
    log_paired_ranges(self.logger, paired_ranges)

    return paired_ranges



# coding: utf-8
"""
    Baseline.Debug.py

    Copyright (c) 2020-2022, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
from matplotlib.patches import Polygon
from MeasuredData import MeasuredData
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from CurveDecomposer import decompose
from LPM import LPM_3d
# from molass_legacy.Baseline.LpmProxy import LpmProxy as LPM_3d
USING_LPM_PROXY = False
from Extrapolation.ExSolver import ExSolver
from molass_legacy.KekLib.NumpyUtils import np_savetxt
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve

BUG_FIX_WITH_REDRAW = False
TITLE_FONTSIZE = 16

def demo(root, in_folder, xray=True, show_preview=False):
    print(in_folder)
    md = MeasuredData(in_folder)

    if show_preview:
        from molass_legacy.Tools.ThreeDimViewer import ThreeDimViewer
        dialog = ThreeDimViewer(root, md)
        dialog.show()

    if xray:
        rd = md.xr
        ecolor = 'orange'
    else:
        rd = md.uv
        ecolor = 'blue'

    rd.set_elution_curve()
    i_slice = rd.i_slice
    j_slice = rd.j_slice
    j_start = j_slice.start
    if j_start is None:
        j_start = 0

    print('i_slice=', i_slice)
    print('j_slice=', j_slice)

    whole_data = rd.data
    data = whole_data[i_slice,j_slice]
    print('data.shape=', data.shape)

    q = rd.vector[i_slice]
    q1 = rd.vector[rd.e_index]

    fig = plt.figure(figsize=(21, 11))
    fig.suptitle("Inspection of Deviation in the Wide Angle Region", fontsize=24)

    gs = GridSpec(2,3)
    ax00 = fig.add_subplot(gs[0,0], projection='3d')
    ax01 = fig.add_subplot(gs[0,1])
    ax02 = fig.add_subplot(gs[0,2])
    ax10 = fig.add_subplot(gs[1,0], projection='3d')
    ax11 = fig.add_subplot(gs[1,1])
    ax12 = fig.add_subplot(gs[1,2])
    z = rd.e_curve.y[j_slice]
    pt_i = rd.e_curve.primary_peak_i - j_start
    hw = 5
    top_slice = slice(pt_i-hw, pt_i+hw)
    range_slice = top_slice
    if USING_LPM_PROXY:
        mapper = md.create_mapper(jslice=j_slice)
    else:
        mapper = None

    i_start = i_slice.start
    if i_start is None:
        i_start = 0

    e_index = rd.e_index - i_start
    demo_impl(ax00, ax01, ax02, data, q, q1, z, top_slice, range_slice, e_index, ecolor, mapper)

    j_slice = slice(846, 1185)
    j_start = j_slice.start
    if j_start is None:
        j_start = 0

    print('j_slice=', j_slice)

    data = whole_data[i_slice,j_slice]
    z = rd.e_curve.y[j_slice]
    pt_i = rd.e_curve.primary_peak_i - j_start
    top_slice = slice(pt_i-hw, pt_i+hw)
    range_slice = slice(900-j_start, 1016-j_start)
    if USING_LPM_PROXY:
        mapper = md.create_mapper(jslice=range_slice)
    else:
        mapper = None

    demo_impl(ax10, ax11, ax12, data, q, q1, z, top_slice, range_slice, e_index, ecolor, mapper, save=False)

    x = rd.vector
    for ax, j0 in [(ax00, 0), (ax10, j_slice.start)]:
        for y in [j_slice.start, j_slice.stop]:
            y_ = y - j0
            ax.plot(x[[0,-1]], [y_, y_], [0, 0], color='yellow' )

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()

def demo_impl(ax00, ax01, ax02, data, q, q1, z, top_slice, range_slice, e_index, ecolor, mapper, save=False):
    ax00.set_title("3D Data View", y=1.09, fontsize=TITLE_FONTSIZE)
    simple_plot_3d(ax00, data, x=q)

    size = data.shape[1]
    x = np.ones(size)*q1
    y = np.arange(size)

    ax00.plot(x, y, z, color=ecolor)

    # raw pt average
    pt_y = np.average(data[:,top_slice], axis=1)

    if save:
        save_data = np.array([q, pt_y, pt_y*0.03]).T
        np_savetxt('temp.dat', save_data)

    # lpm pt average
    if USING_LPM_PROXY:
        lpm = LPM_3d(data, ecurve_y=z, integral=False, for_all_q=True, e_index=e_index, q=q, mapper=mapper)
    else:
        lpm = LPM_3d(data, ecurve_y=z, integral=False, for_all_q=True, e_index=e_index)

    lp_y = np.average(lpm.data[:,top_slice], axis=1)

    M0 = data[:,range_slice]
    M1 = lpm.data[:,range_slice].copy()

    e_curve = ElutionCurve(z)
    lpm.adjust_with_mf(e_index, e_curve)
    lp_y2 = np.average(lpm.data[:,top_slice], axis=1)
    M2 = lpm.data[:,range_slice].copy()
    # solver
    solver = ExSolver()
    c = z[range_slice]
    c = c/np.max(c)

    # raw pt extrapolation (cd=1)
    ex_y0 = solver.solve(M0, c=c, cd=1)

    # lpm pt extrapolation (cd=1)
    ex_y1 = solver.solve(M1, c=c, cd=1)
    ex_y2 = solver.solve(M2, c=c, cd=1)

    ax01.set_title("Peak Top Averages", fontsize=16)
    ax01.set_yscale('log')
    ax01.plot(q, pt_y, label="raw average")
    ax01.plot(q, lp_y, label="lpm average")
    ax01.plot(q, lp_y2, label="lpm+MF average")
    ax01.legend()

    ax02.set_title("Ascending-side Extrapolations", fontsize=16)
    ax02.set_yscale('log')
    ax02.plot(q, ex_y0, label="raw extrapolation (cd=1)")
    ax02.plot(q, ex_y1, label="lpm extrapolation (cd=1)")
    ax02.plot(q, ex_y2, label="lpm+MF extrapolation (cd=1)")
    ax02.legend()



# coding: utf-8
"""
    UvCorrectorTestUtils.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
from time import sleep
from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings
from SerialTestUtils import prepare_serialdata_env
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.UV.PlainCurve import make_secondary_e_curve_at
from molass_legacy.UV.Absorbance import Absorbance
from .PeakRegion import PeakRegion
from .UvCorrectorEgh import UvCorrectorEgh

class Manipulator:
    def __init__(self, parent, logger, canvas_id=1, restart_str=None):
        self.parent = parent
        self.logger = logger
        self.canvas_id = canvas_id
        self.counter = -1
        self.restart_str = restart_str
        self.restarting = restart_str is not None
        self.prepare_out_folder()

    def prepare_out_folder(self):
        self.out_folder = "temp/figs"
        if not os.path.exists(self.out_folder):
            from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
            mkdirs_with_retry(self.out_folder)

    def run_for_all(self):
        from DataUtils import get_pytools_folder, serial_folder_walk
        pytools = get_pytools_folder()
        root_folder = os.path.join(pytools, "Data")
        serial_folder_walk(root_folder, self.do_for_a_file)

    def do_for_a_file(self, in_folder, uv_folder, plot):
        # from TkTester import TestClient  can cause "RuntimeError: main thread is not in main loop"
        clear_temporary_settings()

        try:
            assert in_folder == uv_folder

            self.counter += 1
            if self.restarting:
                if in_folder.find(self.restart_str) >= 0:
                    self.restarting = False
                else:
                    print("skipping", in_folder)
                    return True, None

            sd = prepare_serialdata_env(in_folder, uv_folder)
            absorbance = Absorbance( sd.lvector, sd.conc_array, sd.xray_curve, col_header=sd.col_header )
            data = absorbance.data
            vector = absorbance.wl_vector
            a_curve = absorbance.a_curve
            a_curve2 = make_secondary_e_curve_at(data, vector, a_curve, self.logger)
            fig_file = os.path.join(self.out_folder, "fig-%03d" % self.counter)
            peak_region = PeakRegion(sd.xray_curve, a_curve, a_curve2)
            uvc = UvCorrectorEgh(a_curve, a_curve2)
            uvc.fit(peak_region, debug=True, fig_file=fig_file)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "do_for_a_file(%s) failure: " % in_folder)

        return True, None



"""
    Models.Stochastic.SdmTimeTutorial.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt

def illust_and_proof():
    from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf
    from molass_legacy.Models.Stochastic.Monopore import monopore_func
    print("tutorial")

    x = np.arange(-1000, 800)
    poresize = 100
    rgs = np.array([100, 50, 40, 30, 20])
    rhov = rgs/poresize
    N = 1000
    T = 1
    me = 1.5
    mp = 1.5
    t0 = 100
    tinj = x[0]
    N0 = 3600
    scale = 1
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
    fig.suptitle("SDM Elution Illustration in Different Time Axes with Poresize=%g, $N_0=%d$" % (poresize, N0), fontsize=20)
    for k, (rg, rho) in enumerate(zip(rgs, rhov)):
        ni = N*(1 - rho)**me
        ti = T*(1 - rho)**mp
        for ax, x_ in zip((ax1, ax2), (x, x-tinj)):
            cy = dispersive_monopore_pdf(x - tinj, ni, ti, N0, t0 - tinj, timescale=0.25)
            bop = ">=" if k == 0 else "="
            ax.plot(x_, cy, label="$R_g$%s%g" % (bop, rg))
            if k > 0:
                cy = monopore_func(x, scale, N, T, t0, me, mp, poresize, rg)
                ax.plot(x_, cy,":", color="C%s" % k)
    for ax, tinj_, t0_ in [(ax1, tinj, t0), (ax2, 0, t0 - tinj)]:
        ax.set_title("$t_{inj}=%g$, $t_0=%g$" % (tinj_, t0_), fontsize=16)
        ax.axvline(x=tinj_, color="gray", label="$t_{inj}$")
        ax.axvline(x=t0_, color="red", label="$t_0$")
        hy = np.percentile(ax.get_ylim(), 60)
        dx = t0_ - tinj_
        ax.arrow(x=tinj_, y=hy, dx=dx, dy=0, width=0.003, head_width=0.008,
                    head_length=0.1*dx, length_includes_head=True, color='pink', alpha=0.5)
        ax.legend(loc="upper right")
    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys
    import seaborn as sns
    sns.set_theme()
    sys.path.append("../lib")

    illust_and_proof()


"""
    OurGini.py

    from https://github.com/oliviaguest/gini/blob/master/gini.py
"""
import numpy as np

def gini(array):
    """Calculate the Gini coefficient of a numpy array."""
    # based on bottom eq:
    # http://www.statsdirect.com/help/generatedimages/equations/equation154.svg
    # from:
    # http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm
    # All values are treated equally, arrays must be 1d:
    array = array.flatten()
    if np.amin(array) < 0:
        # Values cannot be negative:
        array -= np.amin(array)
    # Values cannot be 0:
    array += 0.0000001
    # Values must be sorted:
    array = np.sort(array)
    # Index per array element:
    index = np.arange(1,array.shape[0]+1)
    # Number of array elements:
    n = array.shape[0]
    # Gini coefficient:
    return ((np.sum((2 * index - n  - 1) * array)) / (n * np.sum(array)))



"""
    LrfResultPool.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import logging
from molass_legacy._MOLASS.SerialSettings import get_setting

LRF_INFO = 4
USE_LRF_SOLVER = True

class LrfResultPool:
    def __init__(self, pdata, popts):
        self.logger = logging.getLogger(__name__)
        self.lrf_bound_correction = get_setting("lrf_bound_correction")
        self.pdata = pdata
        self.popts = popts
        self.rank_control = get_setting('rank_control')
        self.conc_dependence = get_setting('conc_dependence')

    def run_solver(self, parent=None, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.Extrapolation.PeakSetSelector
            reload(molass_legacy.Extrapolation.PeakSetSelector)  
        from molass_legacy.Extrapolation.PeakSetSelector import PeakSetSelector

        if USE_LRF_SOLVER:
            if debug:
                from importlib import reload
                import molass_legacy.LRF.LrfSolver
                reload(molass_legacy.LRF.LrfSolver)       
            from .LrfSolver import LrfSolver as Solver
        else:
            from molass_legacy.Extrapolation.ExtrapolationSolver import ExtrapolationSolver as Solver

        self.logger.info('cnv_ranges=%s, num_ranges=%s', str(self.pdata.cnv_ranges), str(self.pdata.num_ranges))

        self.solver_results = None
        self.solver = Solver(self.pdata, self.popts)
        self.pno_map = self.solver.get_pno_map()
        self.selector = PeakSetSelector(self.pdata.cnv_ranges, self.pdata.mapper.x_curve, debug=debug)

        if self.rank_control:
            ret_tuple = run_lrf_solver_impl(self.solver, self.selector, lrf_rank=1)
            self.rank1_results = ret_tuple[0]
            ret_tuple = run_lrf_solver_impl(self.solver, self.selector, lrf_rank=2)
            self.rank2_results = ret_tuple[0]
        else:
            lrf_rank = self.conc_dependence
            ret_tuple = run_lrf_solver_impl(self.solver, self.selector, lrf_rank=lrf_rank)
            self.rankx_results = ret_tuple[0]

        self.peak_range_infos = ret_tuple[1]
        self.get_better_results()

    def ok( self ):
        return self.solver_results is not None

    def show_dialog(self, parent, editor_frame=None, last_change_id=None, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Extrapolation.ExtrapolSolverDialog
            reload(molass_legacy.Extrapolation.ExtrapolSolverDialog)
        from molass_legacy.Extrapolation.ExtrapolSolverDialog import ExtrapolSolverDialog
        self.dialog = ExtrapolSolverDialog(parent, self, editor_frame=editor_frame, last_change_id=last_change_id)
        self.dialog.show()

    def get_dialog( self ):
        return self.dialog

    def get_solver_ret_tuple( self ):
        return self.solver_results, self.peak_range_infos

    def get_better_results(self):

        if self.solver_results is None:
            if self.rank_control:
                rank_args = self.get_better_rank_args()
                solver_results = []

                for rank_pair, rank1_result, rank2_result in zip(rank_args, self.rank1_results, self.rank2_results):
                    if rank_pair == 1:
                        solver_results.append(rank1_result)
                    else:
                        solver_results.append(self.synthesize(rank2_result, rank1_result))
            else:
                solver_results = self.rankx_results

            self.solver_results = solver_results

        return self.solver_results

    def get_better_rank_args(self):
        self.peak_judgements = []
        rank_args = []
        for k, range_info in enumerate(self.peak_range_infos):
            num_ad_ranges = len(range_info[2].get_fromto_list())
            print([k], range_info, num_ad_ranges)
            if num_ad_ranges == 1:
                rank_args.append(self.get_minor_rank_info(k, range_info))
            else:
                rank_args.append(self.get_major_rank_info(k, range_info))
        return rank_args

    def get_minor_rank_info(self, k, range_info):
        pno = range_info[0]
        if pno < len(self.peak_judgements):
            pass
        else:
            self.peak_judgements.append(1)
        return self.peak_judgements[pno]

    def get_major_rank_info(self, k, range_info):
        pno, ad = range_info[0:2]
        if pno < len(self.peak_judgements):
            assert ad == 1
        else:
            assert ad == 0

            asc_rank1_result = self.rank1_results[k]
            dsc_rank1_result = self.rank1_results[k+1]
            asc_rank2_result = self.rank2_results[k]
            dsc_rank2_result = self.rank2_results[k+1]

            if self.lrf_bound_correction or True:
                color = self.pno_map.get_color(pno)
                lrf_rank = 2 if color in ["yellow", "red"] else 1
                self.logger.info("[%d-%d] color=%s, lrf_rank=%d", pno, k, color, lrf_rank)
            else:
                asc_Rg1 = asc_rank1_result[LRF_INFO].Rg
                dsc_Rg1 = dsc_rank1_result[LRF_INFO].Rg
                asc_Rg2 = asc_rank2_result[LRF_INFO].Rg
                dsc_Rg2 = dsc_rank2_result[LRF_INFO].Rg

                rg_list = tuple([str(rg) if rg is None else "%.3g" % rg for rg in [asc_Rg1, dsc_Rg1, asc_Rg2, dsc_Rg2]])
                print([k], "(%s, %s), (%s, %s)" % rg_list)

                def judge_as_same(rg1, rg2, allow=0.02):
                    if rg1 is None or rg2 is None:
                        return False

                    ratio = abs((rg1-rg2)/(rg1+rg2))
                    return ratio < allow

                if judge_as_same(asc_Rg1, asc_Rg2, allow=0.01) and judge_as_same(dsc_Rg1, dsc_Rg2, allow=0.01):
                    lrf_rank = 1
                    how = "safely"
                else:
                    if judge_as_same(asc_Rg2, dsc_Rg2, allow=0.02):
                        lrf_rank = 2
                        how = "safely"
                    else:
                        lrf_rank = 1
                        how = "unsurely"

                self.logger.info("LRF rank has been %s judged as %d for the %d-th peak.", how, lrf_rank, pno)

            self.peak_judgements.append(lrf_rank)
        return self.peak_judgements[pno]

    def synthesize(self, rank2_result, rank1_result):
        return rank2_result

def run_lrf_solver_impl(solver, selector, lrf_rank=None,
            exe_queue=None, stop_event=None ):

    results = []
    peak_range_infos = []
    row = 0
    if exe_queue is not None:
        exe_queue.put( [0, row ] )
        if stop_event.is_set():
            return

    for pno, paired_range in enumerate(solver.cnv_ranges):
        fromto_list = paired_range.get_fromto_list()
        for rno, range_ in enumerate(fromto_list):
            start   = range_[0]
            stop    = range_[1] + 1
            peakset_info = selector.select_peakset(row)
            try:
                if USE_LRF_SOLVER:
                    A, B, Z, E, lrf_info, C = solver.solve_range(start, stop,
                                                peakset_info, lrf_rank=lrf_rank,
                                                stop_event=stop_event,
                                                paired_range=paired_range,
                                                )
                else:
                    A, B, Z, E, lrf_info, C = solver.extrapolate_wiser( start, stop, peakset_info,
                                        stop_event=stop_event,
                                        lrf_rank=lrf_rank,
                                        # lrf_rank=None,
                                        )
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb)
                if exe_queue is not None:
                    exe_queue.put( [-1, str(etb) ] )
                raise RuntimeError('in the solver')

            if stop_event is not None and stop_event.is_set():
                print( "solver stopped by stop_event." )
                return None

            results.append( [A, B, Z, E, lrf_info, C] )
            peak_range_infos.append( [pno, rno, paired_range] )
            row += 1
            if exe_queue is not None:
                exe_queue.put( [0, row ] )

    print( "solver finished." )
    return results, peak_range_infos



"""
    HdcTheory.SpikeDemo.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
sns.set_theme()

def demo():
    poresize = 75
    rv = np.linspace(10, poresize, 100)
    rhov = rv/poresize
    rhov[rhov > 1] = 1
    m = 3
    N = 1000
    T = 1
    t0 = 50
    trv = t0 + N*T*(1 - rhov)**m

    lrv = np.linspace(poresize-5, 800, 100)
    lamv = lrv/5000
    C = 2.698
    xv = t0/(1 + 2*lamv - C*lamv**2)

    fig, ax = plt.subplots()
    ax.plot(trv, rv, label=r"$t_0 + N T (1 - \rho)^m$")
    ax.axvline(x=t0, color='red', label=r"$t_0$")
    ax.plot(xv, lrv, label=r"$t_0/(1 + 2 \lambda - C \lambda^2)$")
    ax.set_xlim(0, 300)
    ax.legend()
    fig.tight_layout()
    plt.show()

def valid_range_demo():
    from HdcTheory.ElutionCurve import C
    rv = np.linspace(-200, 1000, 100)
    r0 = 1000
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
    fig.suptitle("Valid $R_{eff}$ Range of $K_{HDC}$ when $R_0$=%g and C=%g" % (r0, C), fontsize=20)
    lamb = rv/r0
    y = 1 + 2*lamb - C*lamb**2
    ax1.set_xlabel(r"$R_{eff}$")
    # Rh/R0 * C= 1
    # Rh = R0/C
    ax1.set_ylabel(r"$1 + 2 \lambda - C \lambda^2$")
    boundary = r0/C
    negative = rv <= 0
    positive = rv > 0
    invalid = rv > boundary
    valid = np.logical_and(positive, rv <= boundary)
    ax1.plot(rv[valid], y[valid])
    ax1.plot(rv[negative], y[negative], ":", color='gray', alpha=0.5)
    ax1.plot(rv[invalid], y[invalid], ":", color='gray', alpha=0.5)
    boundary_label = r"$\lambda=\frac{R_{eff}}{R_0}=\frac{1}{C}$"
    ax1.axvline(x=r0/C, color='pink', label=boundary_label)
    ax1.legend()

    ax2.set_xlabel(r"$K_{HDC}: (1 + 2 \lambda - C \lambda^2)^{-1}$")
    ax2.set_ylabel(r"$R_{eff}$")
    ax2.plot(1/y[valid], rv[valid], label=r"Calibration Curve", color="cyan")
    ax2.plot(1/y[negative], rv[negative], ":", color='gray', alpha=0.5)
    ax2.plot(1/y[invalid], rv[invalid], ":", color='gray', alpha=0.5)
    ax2.axhline(y=r0/C, color='pink', label=boundary_label)
    ax2.axvline(x=1, color='red', label=r"$t_0$")
    # xmin, xmax = ax2.get_xlim()
    ax2.set_xlim(0, 2)
    ax2.legend(loc="upper left")
    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys
    sys.path.append("../lib")    
    # demo()
    valid_range_demo()



"""
    GuinierTools.CpdDecompIndiect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ElutionCurveModels import EGHA
from RgProcess.RgCurveUtils import plot_rg_curve
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from GuinierTools.RgCurveUtils import get_connected_curve_info

def cpd_spike_impl(editor):
    from importlib import reload
    import Peaks.PeakEditorUtils
    reload(Peaks.PeakEditorUtils)
    from molass_legacy.Peaks.PeakEditorUtils import apply_new_peaks
    import GuinierTools.CpdDecompIndirectImpl
    reload(GuinierTools.CpdDecompIndirectImpl)
    from GuinierTools.CpdDecompIndirectImpl import compute_end_points, imporove_decomposition

    print("cpd_spike_impl")
    model = EGHA()
    x, y, peaks, baseline = editor.xr_draw_info
    rg_curve = editor.dsets[1]

    x_, y_, rgv, qualiteis, valid_bools = get_connected_curve_info(rg_curve)

    nc = 4
    end_points = compute_end_points(nc, x_, rgv)
    new_peaks = imporove_decomposition(x, y, model, peaks, end_points)

    in_folder = get_in_folder()

    def plot_decomposition(ax, peaks):
        ax.plot(x, y, color='orange')
        cy_list = []
        for i, peak in enumerate(peaks):
            cy = model(x, peak)
            cy_list.append(cy)
            ax.plot(x, cy, ":", label='component-%d' % (i+1))
        ty = np.sum(cy_list, axis=0)
        ax.plot(x, ty, ":", color='red')
        axt = ax.twinx()
        axt.grid(False)
        plot_rg_curve(axt, rg_curve, label='Rg Curve')
        for i in range(nc):
            x1 = end_points[i]
            x2 = end_points[i+1]
            axt.axvspan(x1, x2, color='C%d'%i, alpha=0.1)
        ymin, ymax = axt.get_ylim()
        axt.set_ylim(15, ymax)
        ax.legend()
        axt.legend(loc='center right')

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("Better Decomposition using CPD of Rg-curve on %s" % in_folder, fontsize=20)
        ax1.set_title("Initial Decomposition", fontsize=16)
        ax2.set_title("Improved Decomposition", fontsize=16)
        plot_decomposition(ax1, peaks)
        plot_decomposition(ax2, new_peaks)
        fig.tight_layout()
        ret = plt.show()
    
    if ret:
        print("apply")
        apply_new_peaks(editor, new_peaks, debug=False)